from virtual.v_size import create_virtual_column_layout, _create_regression
import pytest
from virtual.utils import ModelType, _read_json
from .utils import extract_functions

@pytest.fixture
def model_types():
  return list(map(ModelType, ['sparse-lr']))

@pytest.fixture
def sample_data1():
  return _read_json('tests/sample_data1.json')

@pytest.fixture
def schema1():
  return _read_json('tests/schema1.json')

def test_create_regression(sample_data1, schema1, model_types):
  fns = extract_functions(sample_data1, model_types[0])
  for index, iter in enumerate(sample_data1):
    ret = _create_regression(
      iter['models']['sparse-lr'],
      schema=schema1,
      target_name=iter['target_name']
    )
    other_cols = fns[index].split('=')[1].strip()
    assert ret == f'"{other_cols}"'

def test_create_virtual_column_layout(sample_data1, schema1, model_types):
  for _, iter in enumerate(sample_data1):
    ret = create_virtual_column_layout(
      con=None,
      target_iter=iter,
      schema=schema1,
      model_type=model_types[0]
    )
    col_name = iter['target_name']

    # Create the regression.
    tmp = _create_regression(
      iter['models']['sparse-lr'],
      schema=schema1,
      target_name=col_name
    )
    assert ret == [f'"{col_name}_offset" = coalesce(round("{col_name}" - ({tmp}), 0), 0)']


from sklearn.linear_model import LinearRegression, Ridge
import pyarrow.parquet as pq
import pyarrow as pa
import pyarrow.parquet as pq
import os
import numpy as np
from models.sparse_model import SparseLR
from sklearn.metrics import root_mean_squared_error

# from sklearnex import patch_sklearn
# patch_sklearn()

# (group_size, num_iterations).
k_regression_settings = {2: 10, 4: 10, 8: 10, 16: 10}

# Define which model to use
# MODEL_TYPE = 'linear'
MODEL_TYPE = 'sparse'
# MODEL_TYPE = 'ridge'

def BUILD_MODEL(max_mse_allowed=None):
    if MODEL_TYPE == 'ridge':
        return Ridge(alpha=10.0)
    elif MODEL_TYPE == 'linear':
        return LinearRegression()
    elif MODEL_TYPE == 'sparse':
        assert max_mse_allowed is not None
        return SparseLR(max_mse_allowed)
    else:
        assert 0

def compute_error(model, X, y):
    return root_mean_squared_error(y_true=y, y_pred=model.predict(X)) * X.shape[0]

class K_Regression:
    def __init__(self, num_groups, iterations, max_mse_allowed):
        self.num_groups = num_groups
        self.iterations = iterations
        self.max_mse_allowed = max_mse_allowed

    def initialize(self, X, y):
        permutation = np.random.permutation(X.shape[0])
        X_groups = np.array_split(X[permutation], self.num_groups)
        y_groups = np.array_split(y[permutation], self.num_groups)

        models = [BUILD_MODEL(self.max_mse_allowed).fit(X_groups[i], y_groups[i]) for i in range(self.num_groups)]
        errs = [compute_error(models[i], X_groups[i], y_groups[i]) for i in range(self.num_groups)]
        return models, sum(errs) / X.shape[0]

    def expectation(self, X, y, models):
        # TODO: But is this really the distance to the line?
        distance_matrix = np.asarray([np.abs(models[i].predict(X) - y) for i in range(len(models))])

        # Take the min.
        return np.argmin(distance_matrix, axis=0)

    def maximisation(self, X, y, membership, curr_num_groups):
        models = []

        # TODO: Compute the error and take the best shape.
        errs = []
        for i in range(curr_num_groups):
            # Take the indices.
            indices = np.where(membership == i)[0]

            # Skip if none.
            if indices.size == 0:
                continue

            # Set the data.
            y_i = y[indices]
            X_i = X[indices]

            # Build.
            reg = BUILD_MODEL(self.max_mse_allowed).fit(X_i, y_i)

            # Calculate the weighted error for this slice.
            errs.append(compute_error(reg, X_i, y_i))

            # And append the model.
            models.append(reg)
        
        # Return the weighted error.
        return models, sum(errs) / X.shape[0]

    def fit(self, X, y):
        # Init.
        models, best_err = self.initialize(X=X, y=y)

        best_models = models.copy()
        errs = []
        x = 0
        while True:
            # If there's absolutely no chance to decrease.
            if x == 3 and errs[-1] > 10.0:
                return best_models
            
            # Check the number of iterations.
            x += 1
            if x > self.iterations or self.iterations > 50:
                return best_models
            
            # Expectation.
            membership = self.expectation(X=X, y=y, models=models)

            # Maximization.
            models, curr_err = self.maximisation(X=X, y=y, membership=membership, curr_num_groups=len(models))

            # Append the error.
            errs.append(curr_err)

            # Just stop.
            if curr_err < 1e-4:
                return models

            if curr_err < best_err:
                best_err = curr_err
                best_models = models.copy()
            else:
                # It got worse -> stuck.
                # TODO: Maybe redo????
                return best_models

            # Should we give it a last chance?
            if x == self.iterations - 1 and errs[-1] < 1e-1:
                if errs[-2] - errs[-1] > 1e-3:
                    self.iterations += 5
        return best_models

    # def visualise(self, X, y, models, y_cutoff):
    #     membership = self.expectation(X=X, y=y, models=models)
    #     index = np.where(y < y_cutoff)[0]
    #     X_subset = X[index]
    #     y_subset = y[index]
    #     membership_subset = membership[index]

    #     clrs = sns.color_palette('tab10', n_colors=len(models))

    #     for i in range(self.num_groups):
    #         indices = np.where(membership_subset == i)[0]
    #         y_i = y_subset[indices]
    #         X_i = X_subset[indices].reshape(-1, 1)
    #         plt.scatter(X_i, y_i, color=clrs[i], s=5, alpha=0.3)
    #         plt.plot(X, models[i].predict(X), color=clrs[i])
    #     plt.show()

    def storage_cost(self, X, y, models, precision,path_parquet):
        membership = self.expectation(X=X, y=y, models=models)
        residuals = np.zeros_like(y, dtype=float)

        for i in range(self.num_groups):
            indices = np.where(membership == i)[0]
            if len(indices) == 0:
                continue
            y_i = y[indices]
            X_i = X[indices]
            pred = models[i].predict(X_i)

            res = np.round(pred, decimals=precision) - y_i

            # TODO: This is not so correct, since we first need to cast the prediction.
            if precision == 0:
                res = (pred - y_i).astype(int)

            residuals[indices] = res

        if (membership == 0).all():
            pa_table = pa.table({"offset": residuals.flatten()})
        else:
            pa_table = pa.table({"offset": residuals.flatten(), "reg_line": membership.flatten()})

        pq.write_table(pa_table, path_parquet)
        file_size = os.stat(path_parquet).st_size
        os.remove(path_parquet)
        return file_size, residuals, membership

def storage_cost(df, file_name):
    storage = {}
    # data=np.arange(0, df.shape[0])
    for series_name, series in df.items():

        pa_table = pa.table({"ref": series})
        pa.parquet.write_table(pa_table, file_name)
        file_size = os.stat(file_name).st_size
        storage[series_name] = file_size
        os.remove(file_name)
    return storage



from virtual.expr_parser import ExprParser

def test_complex():
  formula = '''
    case when "total_amount_switch" = 0 then 20.98
    when "total_amount_switch" = 1 then round("fare_amount" + "extra" + "mta_tax" + "tip_amount" + "tolls_amount" + 0.3, 2)
    when "total_amount_switch" = 2 then round(0.02 * "trip_distance" + -0.0069 * "PULocationID" + -0.5003 * "DOLocationID" + 0.3795 * "fare_amount" + 0.0133 * "extra" + 0.0022 * "tip_amount" + 124.9925, 2)
    when "total_amount_switch" = 3 then 21.89
    when "total_amount_switch" = 4 then round("fare_amount" + "extra" + "tip_amount" + 0.8, 2)
    when "total_amount_switch" = 5 then round("fare_amount" + "extra" + "mta_tax" + "tip_amount" + "tolls_amount" + 0.3, 2)
    when "total_amount_switch" = 6 then round("fare_amount" + "extra" + "mta_tax" + "tip_amount" + "tolls_amount" + 0.3, 2)
    when "total_amount_switch" = 7 then 31.42
    when "total_amount_switch" = 8 then round(0.076 * "PULocationID" + -0.2071 * "DOLocationID" + 0.038 * "fare_amount" + -0.0009 * "extra" + 0.0074 * "tip_amount" + 73.5461, 2)
    when "total_amount_switch" = 9 then 10.71
    when "total_amount_switch" = 10 then "fare_amount" + "extra" + "mta_tax" + "tip_amount" + "improvement_surcharge"
    when "total_amount_switch" = 11 then round("fare_amount" + "extra" + "tip_amount" + "tolls_amount" + 0.8, 2)
    else "fare_amount" + "extra" + "mta_tax" + "tip_amount" + "tolls_amount" + "improvement_surcharge" end + "total_amount_offset"
  '''
  # parser = ExprParser(formula)
  # parser.parse()
  assert 1


import math

def debug(d):
  # print(f'[{d["type"]}] --> {d}')
  return d

class ExprParser:
  def __init__(self, expr):
    self.expr_ = expr
    self.pos = 0
    self.mapping = dict()

  def curr_char(self):
    while self.pos != len(self.expr_) and self.expr_[self.pos] == ' ':
      self.pos += 1
    if self.pos == len(self.expr_):
      return None
    return self.expr_[self.pos]

  def curr_token(self):
    # TODO: Maybe store state to the `pos`.
    while self.pos != len(self.expr_) and self.expr_[self.pos] == ' ':
      self.pos += 1

    ahead_pos = self.pos
    while ahead_pos != len(self.expr_) and (self.expr_[ahead_pos] != ' ' and self.expr_[ahead_pos] != ',' and self.expr_[ahead_pos] != '(' and self.expr_[ahead_pos] != ')'):
      ahead_pos += 1
    init_pos = ahead_pos
    while ahead_pos != len(self.expr_) and self.expr_[ahead_pos] == ' ':
      ahead_pos += 1

    self.mapping[self.pos] = ahead_pos
    token = self.expr_[self.pos : init_pos]
    return token
  
  def curr_coef(self):
    # TODO: Maybe store state to the `pos`.
    ahead_pos = self.pos
    while ahead_pos != len(self.expr_) and (self.expr_[ahead_pos].isdigit() or self.expr_[ahead_pos] == '.'):
      ahead_pos += 1
    init_pos = ahead_pos
    while ahead_pos != len(self.expr_) and self.expr_[ahead_pos] == ' ':
      ahead_pos += 1

    self.mapping[self.pos] = ahead_pos
    coef = float(self.expr_[self.pos : init_pos])
    return coef

  def curr_literal(self):
    # TODO: Maybe store state to the `pos`.
    assert self.curr_char() == '"'
    ahead_pos = self.pos + 1
    while ahead_pos != len(self.expr_) and self.expr_[ahead_pos] != '"':
      ahead_pos += 1
    ahead_pos += 1
    init_pos = ahead_pos
    while ahead_pos != len(self.expr_) and self.expr_[ahead_pos] == ' ':
      ahead_pos += 1

    self.mapping[self.pos] = ahead_pos
    literal = self.expr_[self.pos : init_pos]
    return literal

  def advance_char(self):
    self.pos += 1

  def consume(self):
    assert self.pos in self.mapping
    self.pos = self.mapping[self.pos]
    self.mapping.pop(self.pos, None)

  def advance_token(self):
    self.consume()
  
  def advance_coef(self):
    self.consume()

  def advance_literal(self):
    self.consume()

  def parse(self):
    return self.E()

  def when_expr(self):
    # print(f'[when]')
    assert self.curr_token() == 'when'
    self.advance_token()
    cond = self.E()
    assert self.curr_token() == 'then'
    self.advance_token()
    then_stmt = self.E()
    return {
      'type' : 'when',
      'cond' : cond,
      'then' : then_stmt
    }

  def case_expr(self):
    # print(f'[case]')

    assert self.curr_token() == 'case'
    self.advance_token()
    assert self.curr_token() == 'when'

    when_stmts = []
    while self.curr_token() != 'else':
      when_stmts.append(self.when_expr())

    # Parse the end.
    self.advance_token()

    # Parse the `else`.
    else_stmt = self.E()

    assert self.curr_token() == 'end'
    self.advance_token()

    return debug({
      'type' : 'case',
      'when' : when_stmts,
      'else' : else_stmt
    })
  
  def round_expr(self):
    # print(f'[round] {self.expr_[self.pos:]}')
    assert self.curr_token() == 'round'
    self.advance_token()
    assert self.curr_char() == '('
    self.advance_char()

    expr = self.E()
    assert self.curr_char() == ','
    self.advance_char()
    scale = self.T()
    self.advance_char()
    return debug({
      'type' : 'round',
      'expr' : expr,
      'scale' : scale
    })

  def coalesce_expr(self):
    # print(f'[coalesce]')

    assert self.curr_token() == 'coalesce'
    self.advance_token()
    assert self.curr_char() == '('
    self.advance_char()

    exprs = []
    exprs.append(self.E())
    while self.curr_char() == ',':
      self.advance_char()
      exprs.append(self.E())
    self.advance_char()
    assert len(exprs) == 2
    return debug({
      'type' : 'coalesce',
      'exprs' : exprs
    })

  # Represents a generic term.
  def T(self):
    # print(f'[term] {self.expr_[self.pos:]}')

    if self.curr_char() == '(':
      self.advance_char()
      expr = self.E()
      self.advance_char()
      return debug({
        'type' : 'term',
        'coef' : 1.0,
        'expr' : expr
      })
    
    # CASE.
    if self.curr_token() == 'case':
      ret = self.case_expr()
      return ret
    
    # COALESCE.
    if self.curr_token() == 'coalesce':
      ret = self.coalesce_expr()
      return ret

    # ROUND.
    if self.curr_token() == 'round':
      ret = self.round_expr()
      return ret

    # NULL.
    if self.curr_token() == 'null':
      self.advance_token()
      return debug({
        'type' : 'literal',
        'val' : 'null'
      })

    # Literal.
    if self.curr_char() == '"':
      literal = self.curr_literal()

      self.advance_literal()
      # print(f'literal={literal}')
      return debug({
        'type' : 'literal',
        'val' : literal
      })

    # Unary minus.
    if self.curr_char() == '-':
      self.advance_char()
      expr = self.T()
      return debug({
        'type' : 'term',
        'coef' : -1.0,
        'expr' : expr
      })
    
    # Parse coefficient.
    if self.curr_char().isdigit():
      coef = self.curr_coef()
      self.advance_coef()

    # Do we have a multiplication?
    if self.curr_char() == '*':
      self.advance_char()
      expr = self.T()
      return debug({
        'type' : 'term',
        'coef' : coef,
        'expr' : expr
      })

    # Otherwise, we have a simple numeric. This is for the intercept or for the round scale.
    return debug({
      'type' : 'literal',
      'val' : coef
    })

  # Represents a `CAST`.
  def C(self):
    # print(f'[cast]')

    expr = self.T()
    if self.curr_char() == ':':
      self.advance_char()
      assert self.curr_char() == ':'
      self.advance_char()
      cast_type = self.curr_token()
      self.advance_token()
      return debug({
        'type' : 'cast',
        'expr' : expr,
        'cast_type' : cast_type
      })
    return expr

  # Represents a generic expression.
  def E(self):
    # print('[expr]')

    # Generic case.
    exprs = [self.C()]
    while self.curr_char() == '+':
      self.advance_char()
      exprs.append(self.C())

    # If we anyway have a single expression, just take that as the root.
    if len(exprs) == 1:
      return exprs[0]
    
    # Otherwise, build a sum.
    return {
      'type' : 'sum',
      'terms' : exprs
    }


def print_expr(tree):
  if tree['type'] == 'cast':
    expr = print_expr(tree['expr'])
    assert expr
    if not expr[-1] in [')', '"']:
      expr = f'({expr})'
    return f"{expr}::{tree['cast_type']}"

  if tree['type'] == 'case':
    whens = '\n'.join(map(print_expr, tree['when']))
    else_stmt = print_expr(tree['else'])
    return f'(case {whens} else {else_stmt} end)'

  if tree['type'] == 'when':
    cond = print_expr(tree['cond'])
    then_stmt = print_expr(tree['then'])
    return f'when {cond} then {then_stmt}'
  
  if tree['type'] == 'coalesce':
    exprs = ', '.join(map(print_expr, tree['exprs']))
    return f'coalesce ({exprs})'
  
  if tree['type'] == 'round':
    return f"round({print_expr(tree['expr'])}, {print_expr(tree['scale'])})"

  if tree['type'] == 'sum':
    exprs = ' + '.join(map(print_expr, tree['terms']))
    return exprs
  
  if tree['type'] == 'or':
    factors = ' or '.join(map(print_expr, tree['factors']))
    return factors

  if tree['type'] == 'term':
    coef = tree['coef']

    # Optimize out an 1.
    if math.isclose(coef, 1.0):
      return f"{print_expr(tree['expr'])}"
    
    # Optimize out an -1.
    if math.isclose(coef, -1.0):
      return f"-{print_expr(tree['expr'])}"
    
    # Cast to an integer.
    if coef.is_integer():
      coef = int(coef)

    # Generic expresison.
    return f"{coef} * {print_expr(tree['expr'])}"

  if tree['type'] == 'literal':
    if isinstance(tree['val'], str):
      if tree['val'].isdigit():
        tmp = float(tree['val'])
        if tmp.is_integer():
          return f"{int(tmp)}"
    else:
      if tree['val'].is_integer():
        return f"{int(tree['val'])}"
    return f"{tree['val']}"
  
  assert 0  



from utils import ModelType
from typing import List
import pandas as pd
import pathlib
import v_size
import sys

def compute_target_sizes(data: pd.DataFrame | pathlib.Path, functions, schema, model_types: List[ModelType], sample_size=10_000):
  assert schema is not None
  assert isinstance(data, (pd.DataFrame, pathlib.Path))

  # Compute the gains for the target columns.
  target_sizes = v_size.compute_sizes_of_target_columns(functions, data, schema, model_types, sample_size)

  # Error?
  if target_sizes is None:
    assert 0

  # And return.
  return target_sizes

def collect_refs(model, model_type: ModelType):
  ref_idxs, ref_names = [], []
  if model_type.is_k_regression():
    for local_model in model['config']:
      for coeff in local_model['coeffs']:
        if coeff['col_index'] not in ref_idxs:
          assert coeff['col_name'] not in ref_names
          ref_idxs.append(coeff['col_index'])
          ref_names.append(coeff['col_name'])
  else:  
    for coeff in model['coeffs']:
      ref_idxs.append(coeff['col_index'])
      ref_names.append(coeff['col_name'])
  return ref_idxs, ref_names

def conflict(e1, e2):
  if e1['right'] == e2['right']:
    return True
  
  # Cyclic: But now only if the left side of `e2` has only one element.
  if e1['right'] in e2['left'] and e2['right'] in e1['left']:
    # TODO: Wait, this was wrong.
    return True
  return False

def run_greedy(hyperedges):
  # Sort in decreasing order.
  sorted_hyperedges = sorted(hyperedges, key=lambda x: -x['gain'])

  chosen = []

  def is_transitive(he1, he2):
    if he1['right'] in he2['left']:
      return True
    return False

  def combine(he1, he2):
    assert is_transitive(he1, he2)
    return {
      'left' : he1['left'],
      'right' : he2['right'],
      'is_fake' : True
    }

  def merge(he):
    chosen.append(he)
    for i in range(len(chosen)):
      side = 0
      test = 0
      if is_transitive(he, chosen[i]):
        test += 1
        side = -1
      if is_transitive(chosen[i], he):
        test += 1
        side = +1

      assert test <= 1
      if side == -1:
        chosen.append(combine(he, chosen[i]))
      elif side == +1:
        chosen.append(combine(chosen[i], he))

  total_gain = 0
  for i in range(len(sorted_hyperedges)):
    flag = True
    for chosen_he in chosen:
      if conflict(sorted_hyperedges[i], chosen_he):
        flag = False
    if flag:
      total_gain += sorted_hyperedges[i]['gain']
      merge(sorted_hyperedges[i])
      # chosen.append(sorted_hyperedges[i])

  final_chosen = []
  for i in range(len(chosen)):
    if 'is_fake' not in chosen[i]:
      final_chosen.append(chosen[i])
  return total_gain, final_chosen

def compute_space_gain(hyperedge, config, model_type):
  hyperedge['gain'] = config['sizes'][model_type.name]['base'] - config['sizes'][model_type.name]['virtual']

def optimize(configs, model_types: List[ModelType]):
  # Inspect the configs and represent them as hyperedges.
  # That is, the left side of the hyperedge is the set of reference columns.
  # The right side is the target column.
  hyperedges = []
  for config in configs:
    for model_type in model_types:
      if model_type.name not in config['models']:
        continue
      target_index = config['target_index']
      target_name = config['target_name']
      model = config['models'][model_type.name]

      curr_col_idxs, _ = collect_refs(model, model_type)

      hyperedges.append({
        'target_index' : target_index,
        'target_name' : target_name,
        'model_type' : model_type.name,
        model_type.name : model,
        'left' : curr_col_idxs,
        'right' : target_index,
      })

      compute_space_gain(hyperedges[-1], config, model_type)
      if hyperedges[-1]['gain'] <= 0:
        hyperedges = hyperedges[:-1]

  import time
  cp1 = time.time()
  # obj_value1, chosen1 = run_optimal(hyperedges)
  cp2 = time.time()
  obj_value2, chosen2 = run_greedy(hyperedges)
  cp3 = time.time()

  # print(f'Optimal: total_gain={obj_value1}, #corrs={len(chosen1)}/{len(hyperedges)}, time={cp2 - cp1}s')
  # print(f'Greedy: total_gain={obj_value2}, #corrs={len(chosen2)}/{len(hyperedges)}, time={cp3 - cp2}s')

  return {
    # 'optimal' : {
    #   'obj-value' : obj_value1,
    #   'chosen' : chosen1,
    #   'time' : cp2 - cp1
    # },
    'greedy' : {
      'obj-value' : obj_value2,
      'chosen' : chosen2,
      'time' : cp3 - cp2
    }
  }


import pandas as pd
import pathlib
import duckdb

# Custom helper scripts.
from virtual.schema_inference.schema_utils import handle_schema
from virtual.utils import build_query, get_csv_null_options

USE_DUCKDB_PARSER = False

class DataParser:
  def __init__(self, data: pd.DataFrame | pathlib.Path, nrows=None):
    self.data = data
    self.nrows = nrows

  def parse(self):    
    # Parse the schema.

    # TODO: Indeed, do we still use `dtype_mapping`?
    self.column_names, dtype_mapping, csv_dialect, has_header, self.type_cns = handle_schema(self.data)

    # Read the CSV with the dtype mapping and date parsing, and without headers
    # TODO: Why does the other need a `dtype_mapping`?
    # TODO: I think this was because we always used SQL to infer the types.
    # TODO: So we can remove that part, also in the schema inference.

    # Specify the column indices we want to parse.
    # TODO: Also remove strings.
    use_col_idxs = [index for index, cn in enumerate(self.column_names) if cn not in self.type_cns['date']]
    if isinstance(self.data, pathlib.Path):
      if USE_DUCKDB_PARSER:
        # Take the selected columns.
        selected_columns = (
          [self.column_names[idx] for idx in use_col_idxs]
          if use_col_idxs else "*"
        )

        # Construct the DuckDB query
        query = build_query(
          select_stmt=', '.join(selected_columns),
          input=self.data,
          header=has_header,
          delim=csv_dialect['delimiter'],
          quotechar=csv_dialect['quotechar'],
          nullstr=get_csv_null_options(),
          sample_size=self.nrows
        )
        
        # Add row limit if specified.
        # NOTE: This time we _really_ have to read the data.
        if self.nrows:
          query += f" LIMIT {self.nrows}"

        # Fetch the `df`.
        df = duckdb.query(query).to_df()
      else:
        df = pd.read_csv(
          self.data,
          names=self.column_names,
          # TODO: There was indeed a problem here.
          # dtype=dtype_mapping,
          # TODO: Removed this, since it's really slow and we don't need date columns anyway.
          # parse_dates=date_columns,
          usecols=use_col_idxs,
          header=has_header,
          delimiter=csv_dialect['delimiter'],
          quotechar=csv_dialect['quotechar'],
          nrows=self.nrows
        )
    elif isinstance(self.data, pd.DataFrame):
      df = self.data
      if use_col_idxs is not None:
        df = df.iloc[:, use_col_idxs]
      
      # Limit the number of rows.
      if self.nrows is not None:
        df = df.head(self.nrows)
    else:
      # Unsupport data source. Unreachable for now.
      assert 0
    
    # Note: Very important. Update the column names.
    self.column_names = df.columns

    # And return the dataframe.
    return df
    
  def compute_valid_column_indices(self, df):
    valid_column_indices = []
    for i, cn in enumerate(df.columns):
      if cn in self.type_cns['date'] or cn in self.type_cns['string'] or cn in self.type_cns['boolean']:
        continue
      valid_column_indices.append(i)
    return valid_column_indices

  def extract_sample(self, df, sample_size=None):
    valid_column_indices = self.compute_valid_column_indices(df)
    valid_column_names = [self.column_names[i] for i in valid_column_indices]

    df_without_null = df.dropna(subset=valid_column_names)

    # Note: If this doesn't hold, then it's a bit problematic for linear regression.
    # This is because this is an indeterminate system.
    if len(df_without_null) < len(valid_column_indices):
      df_without_null = df.fillna(0)

    if sample_size is None:
      sample_size = 1_000

    # Handle special cases.
    if len(df) == 0:
      sample = df_without_null
      ratio = 1
    else:
      ratio = sample_size / len(df_without_null)

    if ratio > 1:
      ratio = 1

    # And sample.
    sample = df_without_null.sample(frac=ratio, replace=False, random_state=0)

    # Return the sample.
    return sample



from virtual.interface import train, to_format, query
import pandas as pd
import numpy as np
from .utils import extract_functions, compare_fns
import os
import duckdb

def sum_data():
  a = np.random.randint(0, 10, size=1000)
  b = np.random.randint(0, 10, size=1000)
  c = a + b
  return pd.DataFrame({
    'a': a,
    'b': b,
    'c': c
  })

def get_model_types():
  return ['sparse-lr']

def test_train_sum():
  model_types = get_model_types()
  data = sum_data()
  ret = train(data, model_types=model_types)
  fns = extract_functions(ret, model_type='sparse-lr')
  gt = ['a = c + -b', 'c = b + a', 'b = -a + c']
  assert compare_fns(fns, gt)

def test_query_sum():
  model_types = get_model_types()
  data = sum_data()
  ret = train(data, model_types=model_types)

  # Save.
  to_format(data, 'test_query_sum.parquet', functions=ret)

  # Check the queries.
  for col in ['a', 'b', 'c']:
    ret = query(f"select sum({col}) as test from read_parquet('test_query_sum.parquet');")
    act = duckdb.query(f'select sum({col}) as test from data').fetchdf()
    assert ret['test'].dtype == act['test'].dtype
    assert list(ret['test'])[0] == list(act['test'])[0]

  # Remove the file
  os.remove('test_query_sum.parquet')



from v_size import \
  _size_impl, \
  _create_base_table, \
  create_copy_csv, \
  create_virtual_table_layout, \
  create_dump_base, \
  create_dump_virtual, \
  debug_base_table, \
  arrow_compress_parquet, \
  reoptimize_virtual_table_layout
import duckdb
import utils
import os
import pandas as pd
import schema_inference
from schema_inference import schema_utils
from v_reconstructor import measure_reconstruction_latency_per_column
import pathlib

def _apply_arrow(tmp_path, final_path):
  arrow_compress_parquet(tmp_path, final_path)

  # Delete the duckdb parquet file.
  utils._delete_file(tmp_path)

  # And return the final path.
  return final_path

def _compress_impl(con, fn, data: pd.DataFrame | pathlib.Path, schema, target_columns, type, apply_arrow=True, data_hash=None):
  assert isinstance(data, (pd.DataFrame, pathlib.Path))
  assert data_hash is not None

  # Fix the paths.
  tmp_path = f'.{type}_duckdb-{data_hash}.parquet'
  final_path = f'.{type}_arrow-{data_hash}.parquet'

  # Flush from duckdb.
  con.execute(fn(con, schema, target_columns, tmp_path))

  if not apply_arrow:
    return {
      'tmp_path' : tmp_path,
      'final_path' : final_path
    }

  # Compress via arrow for better compression ratios.
  arrow_compress_parquet(tmp_path, final_path)

  # Delete the duckdb parquet file.
  utils._delete_file(tmp_path)

  # And return.
  return final_path

# This is used to avoid duplicated code in `compress` and `benchmark`.
def _load_data(data: pd.DataFrame | pathlib.Path, schema, layout, nrows):
  # Open the connection.
  con = duckdb.connect(database=':memory:', read_only=False)

  # Select the target columns the layout has chosen.
  target_columns = layout['greedy']['chosen']

  # Create the base table.
  con.execute(_create_base_table(target_columns, schema))

  # Load the csv.
  con.execute(create_copy_csv(data, schema, nrows=nrows))

  # Return the connection.
  return con, target_columns

def compress(data: pd.DataFrame | pathlib.Path, schema, layout, optimize_layout=True, nrows=None, enforce_base=False):
  # Get the data hash.
  data_hash = utils.get_data_hash(data)
  
  # Read the schema.
  if isinstance(schema, pathlib.Path):
    schema = utils._read_json(schema)
  elif schema is None:
    schema = schema_utils.generate_schema(data)

  # Load the data.
  con, target_columns = _load_data(data, schema, layout, nrows)

  # Create the offset and outlier columns.
  create_virtual_table_layout(con, target_columns, schema)

  # Re-optimize the constant auxiliary columns.
  if optimize_layout:
    reoptimize_virtual_table_layout(con, target_columns)

  debug_base_table(con, 'after virtual table layout')

  # And compress.
  base_path = None
  if enforce_base:
    base_path = _compress_impl(con, create_dump_base, data, schema, target_columns, 'base', apply_arrow=True, data_hash=data_hash)
  virtual_path = _compress_impl(con, create_dump_virtual, data, schema, target_columns, 'virtual', apply_arrow=True, data_hash=data_hash)

  # Close the connection.
  con.close()

  # Return the paths, depending on what `enforce_base` says.
  if enforce_base:
    assert base_path is not None
    return [base_path, virtual_path]
  return virtual_path

def _latency_impl(parquet_type, parquet_path, schema, all_target_columns, target_iter):
  assert 'model_type' in target_iter
  model_type = utils.ModelType(target_iter['model_type'])

  # Add the sizes into the dictionary.
  if 'latencies' not in target_iter:
    target_iter['latencies'] = dict()
  
  # NOTE: This is _not_ an `elif`!
  if model_type.name not in target_iter['latencies']:
    target_iter['latencies'][model_type.name] = dict()

  # Run.
  latency = measure_reconstruction_latency_per_column(parquet_type, parquet_path, schema, all_target_columns, target_iter)

  # And add them.
  target_iter['latencies'][model_type.name][parquet_type] = latency
  pass

def benchmark(bench_type, data: pd.DataFrame | pathlib.Path, schema, layout, optimize_layout=True, nrows=None):
  assert bench_type in ['size', 'latency']

  # Get the data hash.
  data_hash = utils.get_data_hash(data)

  # Read the schema.
  if isinstance(schema, pathlib.Path):
    schema = utils._read_json(schema)
  elif schema is None:
    schema = schema_utils.generate_schema(data)

  # Load the data.
  con, target_columns = _load_data(data, schema, layout, nrows)

  # If we want to measure the latency, duplicate the dataset until we have 1M rows.
  if bench_type == 'latency':
    # Double the size until you reach 1M.
    curr_size = utils._get_size(con)
    while 2 * curr_size <= 1_000_000:
      con.execute(f'insert into base_table select * from base_table;')
      curr_size = utils._get_size(con)

    # Insert until 1M rows if we didn't reach the limit yet.
    if curr_size < 1_000_000:
      con.execute(f'insert into base_table select * from base_table limit {1_000_000 - curr_size};')

    print(f'Loaded with repetition: {utils._get_size(con)}')

  # Finalize the layout. This works for both simple and k-regression.
  # NOTE: The `model_type` is also `None` in this case.
  create_virtual_table_layout(con, target_columns, schema)

  # Re-optimize the constant auxiliary columns.
  if optimize_layout:
    reoptimize_virtual_table_layout(con, target_columns)

  # Compute the base and virtual sizes.
  if bench_type == 'size':
    for target_column in target_columns:
      _size_impl(con, schema, target_column, data_hash=data_hash)

  path_mapping = {
    'base' : _compress_impl(con, create_dump_base, data, schema, target_columns, 'base', apply_arrow=False, data_hash=data_hash),
    'virtual' : _compress_impl(con, create_dump_virtual, data, schema, target_columns, 'virtual', apply_arrow=False, data_hash=data_hash)
  }

  # Close the connection. NOTE: We do this trick just to avoid having the table active in the connection.
  # From now on, we won't need the connection anymore.
  con.close()

  # Now apply arrow.
  for key in path_mapping:
    _apply_arrow(path_mapping[key]['tmp_path'], path_mapping[key]['final_path'])
    path_mapping[key] = path_mapping[key]['final_path']

  # Calculate the sizes.
  base_size = os.path.getsize(path_mapping['base'])
  virtual_size = os.path.getsize(path_mapping['virtual'])

  # Compute the base and virtual sizes.
  for target_column in target_columns:
    # Measure the latencies.
    for parquet_type in ['base', 'virtual']:
      _latency_impl(parquet_type, path_mapping[parquet_type], schema, target_columns, target_column)
        
  # Delete the paths.
  for parquet_type in path_mapping:
    utils._delete_file(path_mapping[parquet_type])
 
  # Close the connection.
  # NOTE: Should have been already closed.
  # con.close()

  # And return the sizes.
  return {
    'base_size' : base_size,
    'virtual_size' : virtual_size,
    'cols' : target_columns
  }


import pandas as pd
import numpy as np
import pathlib
from typing import Optional

# Helper scripts
from data_handler import DataParser

MAX_MSE_ALLOWED = 5

from sklearn.linear_model import LinearRegression
from models.sparse_model import SparseLR
from models.k_regression import K_Regression, k_regression_settings, compute_error
from sklearn.metrics import root_mean_squared_error

def run_model(model_type, X, y, col_name=None):
  model, y_pred = None, None
  if model_type == 'lr':
    model = LinearRegression().fit(X, y)
    y_pred = model.predict(X)
  elif model_type == 'sparse-lr':
    model = SparseLR(max_mse_allowed=MAX_MSE_ALLOWED).fit(X, y)
    y_pred = model.predict(X)
  elif model_type == 'k-regression':
    k_models = []
    for num_groups, num_iterations in k_regression_settings.items():
      model = K_Regression(num_groups=num_groups, iterations=num_iterations, max_mse_allowed=MAX_MSE_ALLOWED)

      # NOTE: This is done on the sample.
      # Similarly, the score is computed on the sample.      
      best_model_config = model.fit(X, y)

      # Recompute the membership.
      membership = model.expectation(X=X, y=y, models=best_model_config)

      global_err = 0
      valid_models = []
      for i in range(len(best_model_config)):
        indices = np.where(membership == i)[0]
        if len(indices) == 0:
          continue

        # Take this model since it has some membership.
        valid_models.append(best_model_config[i])

        # Compute the error.
        local_err = compute_error(best_model_config[i], X[indices], y[indices])
        global_err += local_err

      # Update the global error.
      global_err /= X.shape[0]

      # Register the model.
      k_models.append({
        'mse' : global_err,
        'models' : valid_models
      })
    return k_models

  assert model is not None
  mse = root_mean_squared_error(y_true=y, y_pred=y_pred)
  return mse, model.intercept_, model.coef_

def virtualize_table(data: pd.DataFrame | pathlib.Path, nrows=None, sample_size=None, allowed_model_types: Optional[list[str]]=None):
  assert isinstance(data, (pd.DataFrame, pathlib.Path))
  
  # Parse data.
  parser = DataParser(data, nrows)
  df = parser.parse()

  # Select the valid column indices.
  valid_column_indices = parser.compute_valid_column_indices(df)

  # Analzye the coefficients of the regression.
  def reduce_coeffs(coeffs, valid_column_indices):
    selected = []
    for index, coeff in enumerate(coeffs):
      if abs(coeff) < 1e-6:
        continue
      selected.append({
        'col_index' : valid_column_indices[index],
        'col_name' : parser.column_names[valid_column_indices[index]],
        'coeff' : coeff,
      })
    return selected

  # Unwrap a configuration for a given `k`.
  def unwrap_k_regression_config(config):
    ret = []
    for model in config['models']:
      reduced_coeffs = reduce_coeffs(model.coef_, input_columns)

      # NOTE: Even if there are no selected coefficients, we still have to store the intercept.
      # This means that the column is composed of many constants.
      # See for instances column `improvement_surcharge` in Taxi: it only takes values in {-0.3, 0.3}.
      ret.append({
        'intercept' : model.intercept_,
        'coeffs' : reduced_coeffs
      })
    return ret

  results = []
  if len(valid_column_indices) <= 1:
    print('Only one numerical column.')
    return {}
  
  # TODO: We can try multiple samples.
  sample = parser.extract_sample(df, sample_size=sample_size)
  
  # Convert to numpy.
  df = df.to_numpy()
  sample = sample.to_numpy()

  debug_path = ''
  if isinstance(data, pathlib.Path):
    debug_path = str(data.name)

  # Try each column.
  for target_index in valid_column_indices:
    input_columns = valid_column_indices.copy()
    input_columns.remove(target_index)

    # TODO: Maybe try multiple samples.
    # TODO: This would also be helpful for k-regression.
    X = sample[:, input_columns]
    y = sample[:, target_index]

    # Init the local results.
    local_results = None
    try:
      local_results = {
        'target_index': target_index,
        'target_name': parser.column_names[target_index],
        'target_stats': {
          'mean': y.mean(),
          'max': float(y.max()),
          'min': float(y.min())
        },
        'models': {}
      }
    except Exception as e:
      assert False, f"Please double check your CSV file. There is a type/format inconsistency error on column {parser.column_names[target_index]}."
    assert local_results is not None

    # Try the standard models.
    for model_type in ['sparse-lr']:
      # Check if this model is allowed.
      # NOTE: If `allowed_model_types` is not None, we actually want to run this model!
      if allowed_model_types is not None:
        if model_type not in allowed_model_types:
          continue

      # Run the model.
      err, intercept, coeffs = run_model(model_type, X, y, parser.column_names[target_index])

      # Do we exceed the max. allowed mse?
      if err > MAX_MSE_ALLOWED:
        continue

      # Reduce the coefficients.
      reduced_coeffs = reduce_coeffs(coeffs, input_columns)

      # NOTE: Even if there are no selected coefficients, we still have to store the intercept.
      # This means that the column is simply constant.
      local_results['models'][model_type] = {
        'mse' : err,
        'intercept' : intercept,
        'coeffs' : reduced_coeffs
      }

    # Try k-regression, only if it's actually allowed.
    if allowed_model_types is None or 'k-regression' in allowed_model_types:
      # Run the model.
      k_config = run_model('k-regression', X, y, parser.column_names[target_index])

      # Analyze the result found.
      # It could be that the actual _valid_ models are fewer than the group size we ran for.
      seen_ks = set()
      for iter in k_config:
        actual_k = len(iter['models'])

        # Skip, since we anyway do sparse LR.
        if actual_k == 1:
          continue

        # Do we exceed the max. allowed mse?
        if iter['mse'] > MAX_MSE_ALLOWED:
          continue

        # Already seen and better _or_ new?
        if (actual_k in seen_ks and iter['mse'] < local_results['models'][f'{actual_k}-regression']['mse']) or (actual_k not in seen_ks): 
          local_results['models'][f'{actual_k}-regression'] = {
            'mse' : iter['mse'],
            'config' : unwrap_k_regression_config(iter)
          }
        
        # Mark it as seen anyway.
        seen_ks.add(actual_k)
    
    # Only put once.
    if local_results['models']:
      results.append(local_results)

  print(f'We found {len(results)} function(s) in your table.')
  return results


import os
import duckdb
import pandas as pd
import numpy as np
import pyarrow.parquet as pq
import json
import re
from multiprocessing import Lock
from typing import TypedDict
from typing import List
import pathlib
import hashlib
import csv
import btrblocks_utils
import schema_inference

def to_csv(parquet_path, format_path):
  df = pd.read_parquet(parquet_path)
  df.to_csv(format_path, index=False)
  _delete_file(parquet_path)
  return format_path

def to_btrblocks(parquet_path, format_path, schema=None):
  tmp_path = format_path + '.csv'
  df = pd.read_parquet(parquet_path)
  schema = schema_inference.schema_utils.generate_schema(df)
  df.to_csv(tmp_path, index=False)
  btrblocks_utils.convert_csv_to_btrblocks(tmp_path, format_path, schema)
  os.remove(tmp_path)
  _delete_file(parquet_path)
  return format_path

def custom_round(x):
  # TODO: Maybe -- if the weight is almost an integer, convert it?
  return round(x, 4)

def build_query(select_stmt, input, header, delim=',', quotechar='"', nullstr=[], sample_size=None):
  return f"""
    select {select_stmt}
    from read_csv(
      '{input}',
      header={str(header).lower()},
      delim='{delim}',
      quote='{quotechar}',
      nullstr={nullstr}""" \
    + (f", sample_size={sample_size}" if sample_size is not None else ", sample_size=-1") \
    + ")"

def get_csv_null_options():
  return "['NULL', 'null', '', 'None']"

# Get a hash for this data.
def get_data_hash(data: pd.DataFrame | pathlib.Path):
  assert isinstance(data, (pd.DataFrame, pathlib.Path))

  # Take a fingerprint of the data.
  if isinstance(data, pd.DataFrame):
    # The column names.
    column_string = ','.join(data.columns)
  elif isinstance(data, pathlib.Path):
    # The path itself.
    column_string = str(data.resolve())
  else:
    raise TypeError("Input must be a pandas `DataFrame` or a `pathlib.Path`.")

  # And return the hash.
  return hashlib.md5(column_string.encode('utf-8')).hexdigest()

def infer_dialect(file_path, num_lines=5):
  with open(file_path, 'r', encoding='utf-8') as file:
    lines = [file.readline() for _ in range(num_lines)]

    # Detect delimiters using csv.Sniffer
    sniffer = csv.Sniffer()
    sample = ''.join(lines)
    try:
      dialect = sniffer.sniff(sample)
      return {
        'delimiter' : dialect.delimiter,
        'quotechar' : dialect.quotechar
      }
    except csv.Error:
      # Fallback.
      return get_default_csv_dialect()
    
def get_default_csv_dialect():
  return {
    'delimiter' : ',',
    'quotechar' : '"'
  }

def get_duckdb_conn():
  return duckdb.connect(database=':memory:', read_only=False)

def rename(src, dst):
  os.rename(src, dst)

def add_metadata_pq(parquet_path, info, info_name, new_parquet_path=None):
  table = pq.read_table(parquet_path)
  metadata = table.schema.metadata

  if metadata is None:
    pass
  elif info_name.encode('utf-8') in metadata:
    print('The functions are already present in the metadata.')
    return -1
  functions_str = json.dumps(info)
  custom_metadata = {info_name : functions_str}
  merged_metadata = { **custom_metadata, **(table.schema.metadata or {}) }

  fixed_table = table.replace_schema_metadata(merged_metadata)
  if new_parquet_path is not None:
    pq.write_table(fixed_table, new_parquet_path)
  else:
    pq.write_table(fixed_table, parquet_path)
  return 1

def get_metadata_pq(parquet_path, info_name):
  table = pq.read_table(parquet_path)
  metadata = table.schema.metadata
  if metadata is None:
    print('Nothing in parquet metadata.')
    return None
  elif info_name.encode('utf-8') not in metadata:
    print('The functions are not present the metadata.')
    return None
  metadata = metadata[info_name.encode('utf-8')].decode('utf-8')
  return json.loads(metadata)

def rreplace(text, old, new):
  return new.join(text.rsplit(old, 1))

# Precompile the regex pattern to ensure it ends exactly with "-regression"
k_regression_pattern = re.compile(r'^\d+-regression$')

# Regex for formula, e.g. f0, f1.
formula_pattern = re.compile(r'^f\d+$')

def is_formula(token):
  return bool(formula_pattern.match(token))

class SchemaEntry(TypedDict):
  type: str
  scale: int

class ModelType:
  def __init__(self, name):
    self.name = name

  def is_k_regression(self):
    return bool(k_regression_pattern.match(self.name))

  def extract_k(self):
    assert self.is_k_regression()
    assert len(self.name.split('-')) == 2
    try:
      val = int(self.name.split('-')[0])
      return val
    except Exception as e:
      # This should never happen.
      print(e)
      return None

  def __repr__(self):
    return f"ModelType(name={self.name})"

# Check if the type is an integer.
def is_integer(type):
  return 'int' in type.lower()

# Check if the type is a floting point.
# TODO: Extend with decimal for the future.
def is_double(type):
  if 'decimal' in type.lower():
    assert 0
  return any(x in type.lower() for x in ['double', 'float'])

def is_virtualizable(type):
  return is_integer(type) or is_double(type)

def is_attached_column_of(target_name, column_name):
  # TODO: This was a real hack.
  return column_name.startswith(f'{target_name}_')# or (column_name[1:].startswith(f'{target_name}_')

def collect_lr_column_names(model):
  cols = []
  for iter in model['coeffs']:
    cols.append(iter['col_name'])
  return cols

def _get_column(schema, column_name):
  for iter in schema:
    if iter['name'] == column_name:
      return iter
  return None

def _get_nullable(schema, column_name):
  column = _get_column(schema, column_name)
  return 'null' if column['null']['any'] else 'not null'

def _get_info(schema, column_name):
  column = _get_column(schema, column_name)

  if column['type'] == 'DOUBLE':
    return {'type' : 'DOUBLE', 'scale' : column['scale']}
  return {'type' : column['type'], 'scale' : 0}

def _read_json(json_path):
  # Check if the file exists.
  assert os.path.isfile(json_path)

  # And read the data.
  f = open(json_path, 'r', encoding='utf-8')
  data = json.load(f)
  f.close()

  # Return it.
  return data

def _write_json(json_path, json_content):
  f = open(json_path, 'w', encoding='utf-8')
  json.dump(json_content, f, ensure_ascii=False)
  f.close()

def _delete_file(filepath):
  assert os.path.exists(filepath)
  os.remove(filepath)

def csv_path_to_schema_path(prefix, csv_path):
  basename = os.path.basename(csv_path).replace('.csv', '')
  return f'{prefix}/{basename}.json'

def _get_curr_columns(con):
  return list(set(con.execute("select * from base_table limit 1;").fetchdf().columns))

def _get_size(con):
  return list(con.execute('select count(*) as size from base_table;').fetchdf()["size"])[0]

def _get_parquet_header(con, parquet_path):
  return list(con.execute(f"select distinct path_in_schema from parquet_metadata('{parquet_path}')").df()["path_in_schema"])

def _get_distinct_counts(con):
  curr_columns = _get_curr_columns(con)
  assert curr_columns
  return con.execute(f'select ' + ','.join([f"(count(distinct \"{x}\") + count(distinct case when \"{x}\" is null then 1 end)) as \"{x}_count\"" for x in curr_columns]) + ' from base_table').fetchdf().to_dict()

def dump_data(csv_filename, data, filename, prefix, key=None):
  csv_basename = os.path.basename(csv_filename)
  csv_basename_noext = os.path.splitext(csv_basename)[0]
  json_filename = os.path.join(prefix, f'{filename}_{csv_basename_noext}.json')
  
  with open(json_filename, 'w', encoding='utf-8') as f:
    if key is not None:
      json.dump(sorted(data, key=key), f, ensure_ascii=False, indent=2)
    else:
      json.dump(data, f, ensure_ascii=False, indent=2)
  return

def dump_data_without_filename(data, filename, prefix, key=None):
  json_filename = os.path.join(prefix, f"{filename}.json")
  
  with open(json_filename, 'w', encoding='utf-8') as f:
    if key is not None:
      json.dump(sorted(data, key=key), f, ensure_ascii=False, indent=2)
    else:
      json.dump(data, f, ensure_ascii=False, indent=2)
  return

def dump_json_data(data, json_data, component_name, prefix=None):
  # Only dump if the user wished so.
  if prefix is None:
    return
  
  # Dump (even if the result is empty).
  # First, ensure that the folder exists.
  prefix_folder = pathlib.Path(prefix)
  prefix_folder.mkdir(parents=True, exist_ok=True)
  
  # And dump.
  if isinstance(data, pathlib.Path):
    dump_data(data, json_data, component_name, prefix)
  elif isinstance(data, pd.DataFrame):
    dump_data_without_filename(json_data, component_name, prefix)
  else:
    assert 0

def gather_all_models(functions):
  # Collect all model names present in the json file.

  all_keys = set()
  for col in functions:
    all_keys |= set(col['models'].keys())
  return all_keys

def select_models(model_types: List[str], functions):
  # TODO: Now we need to change, since we always want a functions type.

  # Collect all model names present in the json file.
  all_keys = gather_all_models(functions)

  # Select only those model names that are also present in the json file.
  validated = []
  for model_type in model_types:
    # Already present?
    if model_type in all_keys:
      validated.append(model_type)
    elif model_type == 'k-regression':
      # Otherwise, we check for `k-regression`, since these have a number in front.
      for possible_match in all_keys:
        # Is it a k-regression?
        if ModelType(possible_match).is_k_regression():
          validated.append(possible_match)

  # Map to `ModelType`.
  return validated

def natural_rounding(coeff):
  return round(coeff, 4)

# Maybe we need to implement this:
# https://stackoverflow.com/questions/3951505/how-do-i-round-a-set-of-numbers-while-ensuring-the-total-adds-to-1
def rounding(combs):
  import copy
  new_combs = copy.deepcopy(combs)

  # Round a linear regression model.
  def round_model(model):
    for col in model['coeffs']:
      curr_coeff = col['coeff']
      new_coeff = natural_rounding(curr_coeff)
      col['coeff'] = new_coeff
    model['intercept'] = natural_rounding(model['intercept'])

  # Round all numbers present in the json file.
  for combination in new_combs:
    for model_type in combination['models']:
      if model_type in ['lr', 'sparse-lr']:
        round_model(combination['models'][model_type])
      elif ModelType(model_type).is_k_regression():
        for local_model in combination['models'][model_type]['config']:
          round_model(local_model)
      else:
        assert 0
  return new_combs

def compute_metrics(table, combs):
  def get_metrics(y1, y2):
    assert y1.shape == y2.shape

    # TODO: This should be the precision from decimal, if any!
    diff = abs(y1 - y2)
    diff = np.array([x if x is not pd.NA else np.nan for x in diff], dtype=float)
    diff = diff[~np.isnan(diff)]
    equal = (diff < 1e-6).mean()
    average = diff.mean()
    maximum = float(diff.max())
    return {'equal' : equal, 'avg_diff' : average, 'max_diff': maximum}

  def estimate_y(model):
    intercept = model['intercept']
    coeffs = np.array([entry['coeff'] for entry in model['coeffs']])
    col_indices = [entry['col_index'] for entry in model['coeffs']]

    # Estimate the y-values.
    y_est = intercept + np.dot(table[:, col_indices], coeffs)

    # And return it.
    return y_est

  def find_closest_estimate(ests, target):
    # Convert 'ests' to a NumPy array if it's not already
    ests = np.array(ests)  

    # Compute the absolute difference between each estimate and the target
    diffs = np.abs(ests - target)  
    
    # Find the index of the minimum difference along axis 0 (for each column)
    argmin = np.argmin(diffs, axis=0)  
    
    # Use fancy indexing to select the closest estimates based on the minimum indices
    return ests[argmin, np.arange(argmin.size)]

  metrics = []
  for comb in combs:
    comb['metrics'] = dict()
    for model_type in comb['models']:
      if model_type in ['lr', 'sparse-lr']:
        y_est = estimate_y(comb['models'][model_type])

        # Get the metrics.
        local_metrics = get_metrics(y_est, table[:, comb['target_index']])
        comb['metrics'][model_type] = local_metrics
      elif ModelType(model_type).is_k_regression():
        local_ests = []
        for local_model in comb['models'][model_type]['config']:
          local_ests.append(estimate_y(local_model))

        # Get the estimate closest to table[:, comb['target_index']].
        closest_est = find_closest_estimate(local_ests, table[:, comb['target_index']])
        local_metrics = get_metrics(closest_est, table[:, comb['target_index']])
        comb['metrics'][model_type] = local_metrics
      else:
        assert 0
    metrics.append(comb)
  return metrics

def compute_avg_stats(parquet_file, config_path, model_type='sparse-lr'):
  # Open the connection.
  con = duckdb.connect(database=':memory:', read_only=False)

  # Read the config.
  config = _read_json(config_path)

  avgs = {}
  for col_config in config['greedy']['chosen']:
    compare_value = list(con.execute(f"select avg(\"{col_config['target_name']}\") as res from read_parquet('{parquet_file}');").fetchdf()["res"])[0]
    avgs[col_config['target_name']] = compare_value
  return avgs


from sklearn.linear_model import LinearRegression
from sklearn.metrics import root_mean_squared_error
import numpy as np
# from celer import Lasso
import scipy

# from sklearnex import patch_sklearn
# patch_sklearn()

class SparseLRInference:
  def __init__(self, intercept, coeffs):
    self.intercept_ = intercept
    self.coef_ = np.asarray(coeffs)
    pass

  def predict(self, x):
    return np.dot(x, self.coef_) + self.intercept_

def apply_lasso(x, y, selected):
  new_model = Lasso().fit(x[:, selected], y)
  new_y_pred = new_model.predict(x[:, selected])
  new_mse = root_mean_squared_error(y_true=y, y_pred=new_y_pred)

  # if new_mse > 1e-6:
    # return None
  return new_model.intercept_, new_model.coef_

class SparseLR:
  def __init__(self, max_mse_allowed):
    self.max_mse_allowed = max_mse_allowed
    pass     

  def fit(self, x, y):
    # The initial model.
    init_model = LinearRegression().fit(x, y)
    
    # The prediction.
    y_pred = init_model.predict(x)

    # Compute the original MSE.
    original_mse = root_mean_squared_error(y_true=y, y_pred=y_pred)
    if original_mse > self.max_mse_allowed:
      return init_model

    # Take the non-zero features only.
    selected = []
    for index, coeff in enumerate(init_model.coef_):
      if abs(coeff) < 1e-6:
        continue
      selected.append(index)

    # Is the column constant?
    if not selected:
      return SparseLRInference(init_model.intercept_, [0] * x.shape[1])

    # Fit the first LR model.
    test_model = LinearRegression().fit(x[:, selected], y)
    test_y_pred = test_model.predict(x[:, selected])
    test_mse = root_mean_squared_error(y_true=y, y_pred=test_y_pred)

    # Is it work? Then take the original model.
    # TODO: Sometimes this actually happens.
    if test_mse > original_mse + 1e-6:
      return init_model
    
    # If there is one attribute, then we don't have anything to remove.
    if len(selected) == 1:
      return init_model

    temp_mse = original_mse
    fixed_window_mse = original_mse
    lazy_counter = 0

    # TODO: Maybe we can sort by the error from the previous iteration.
    while len(selected) > 1:
      # Set to false.
      has_removed_feature = False

      # Take each feature.
      for feature_index in selected:
        # Fit
        new_selected = [index for index in selected if index != feature_index]
        curr_model = LinearRegression().fit(x[:, new_selected], y)
        curr_y_pred = curr_model.predict(x[:, new_selected])
        curr_mse = root_mean_squared_error(y_true=y, y_pred=curr_y_pred)

        # Did we improve?
        if curr_mse < temp_mse + 1e-6:
          # Try to avoid sequences of errors that don't lead anywhere.
          lazy_counter += 1
          if lazy_counter == 2:
            # No improve since the last step?
            # Assumption: `fixed_window_mse` is much larger and `temp_mse` decreased compared to it.
            # If the decrease is not that much, try with another attribute.
            if fixed_window_mse - temp_mse < 1e-6:
              # Update the counter.
              lazy_counter -= 1
              continue

            # If the decrease is significantly, update the metadata, as follows:
            # Update the fixed-window mse.
            fixed_window_mse = temp_mse

            # Update the counter.
            lazy_counter = 0

          # Regular path: We remove this feature.
          # Remove this feature.
          selected.remove(feature_index)

          # Mark that we removed.
          has_removed_feature = True

          # Update the error.
          temp_mse = curr_mse

          # Out of the for loop.
          break

      # No feature removed?
      if not has_removed_feature:
        break

    # Fit again.
    best_model = LinearRegression().fit(x[:, selected], y)
    best_y_pred = best_model.predict(x[:, selected])
    # best_mse = root_mean_squared_error(y_true=y, y_pred=best_y_pred)

    best_coeffs = []
    index_in_selected = 0
    for i in range(x.shape[1]):
      if i in selected:
        best_coeffs.append(best_model.coef_[index_in_selected])
        index_in_selected += 1
      else:
        best_coeffs.append(0.0)
    return SparseLRInference(best_model.intercept_, best_coeffs)


import duckdb
import pathlib
import pandas as pd
from virtual.utils import infer_dialect

def format_col_name(col_name):
  if col_name.startswith('"') and col_name.endswith('"'):
    return col_name
  return f'"{col_name}"'

class LWSchemaInferer:
  def __init__(self, data: pd.DataFrame | pathlib.Path):
    assert isinstance(data, (pd.DataFrame, pathlib.Path))
    self.data = data
    
  def infer(self, nrows=None):
    # Start by defining the query
    query = None
    if isinstance(self.data, pathlib.Path):
      # Infer the dialect.
      dialect = infer_dialect(self.data)

      # Specify the query.
      query = f"select * from read_csv('{self.data}', header=true, delim='{dialect['delimiter']}', quote='{dialect['quotechar']}'" \
        + (f", sample_size={nrows}" if nrows is not None else ", sample_size=-1") \
        + ", nullstr=['NULL', 'null', '', 'None'])"
    elif isinstance(self.data, pd.DataFrame):
      duckdb.register("mydf", self.data)
      query = f'SELECT * FROM mydf' + (f'LIMIT {nrows}' if nrows is not None else '')
    else:
      assert 0
      
    # Execute the query.
    result = duckdb.sql(query)
    column_types = result.types
    columns = result.columns

    # Determine if the data has a header
    # TODO: Fix this since might not be reliable.
    if columns[0] != "column0":
      has_header = 0
    else:
      has_header = None

    type_mapping = {columns[i]: str(column_types[i]) for i in range(len(columns))}  
    return type_mapping, has_header



import pathlib
import sys
import os

# Set the path for the modules to load.
base_path = pathlib.Path(__file__).resolve().parent
sys.path.insert(0, str(base_path))

# The system components.
import v_driller
import v_optimizer
import v_compressor
import v_reconstructor

# Other utils.
from typing import Optional
import schema_inference
import pandas as pd
import utils
import re

# TODO: We should add `nrows`. Maybe also the sample_size.
def train(data: pd.DataFrame | pathlib.Path | str, nrows=None, sample_size=10_000, model_types: Optional[list[str]]=None, prefix=None):
  """
    Computes the functions hidden in `data`.

    Args:
      data (pd.DataFrame | pathlib.Path | str): The data source.
      nrows (int): The number of rows to take.
      sample_size (int): The number of rows to train on.
      model_types (list[str]): The models that the user wants to explore. If `None`, then we try all our models (for better file sizes).

      [prefix: The directory which to save the results to; helpful for debugging purposes.]

    Example (1):
      df = pd.read_csv('file.csv')
      functions = virtual.train(df)
  """

  # Convert into an actual `Path`.
  if isinstance(data, str):
    data = pathlib.Path(data)
  assert isinstance(data, (pd.DataFrame, pathlib.Path))

  # Check that we're dealing with a file.
  if isinstance(data, pathlib.Path):
    assert data.is_file()

  # Virtualize the table. Note that for this particular step we don't require a schema.
  functions = v_driller.virtualize_table(data, nrows=nrows, sample_size=sample_size, allowed_model_types=model_types)

  # Dump `functions`.
  utils.dump_json_data(data, functions, 'driller', prefix)

  # And return them.
  return functions

# TODO: Check whether all cases for `nrows` work out.
def to_format(data: pd.DataFrame | pathlib.Path | str, format_path, functions=None, schema=None, nrows=None, model_types: Optional[list[str]]=None, prefix=None):
  """
    Converts `data` to a Parquet file.

    Args:
      data (pd.DataFrame | pathlib.Path | str): The data source.
      format_path (str): The file which to save to; will be created if it doesn't exist.
      functions (dict): The functions to be used during compression. If `None`, then `virtual` automatically computes them.
      schema (dict): The schema of the data. If `None`, then `virtual` automatically infers it.
      nrows (int): The #rows to save to Parquet. If `None`, the _entire_ file will be processed.
      model_types (list[str]): The models that the user wants to explore. If `None`, then all are tried out.

      [prefix: The directory which to save the results to; helpful for debugging purposes.]

    Example (1):
      df = pd.read_csv('file.csv')
      virtual.to_format(df, 'file.parquet')

    Example (2):
      virtual.to_format('file.csv', 'file.parquet')
  """

  # Check the format path.
  assert '.' in format_path, 'The `format_path` is invalid!'
  format_type = os.path.splitext(format_path)[-1]
  format_type = format_type[1:]
  assert format_type, 'The `format_path` is invalid!'
  if format_type not in ['csv', 'parquet', 'btrblocks']:
    assert False, f'Format {format_type[1:]} not yet supported!'

  # Convert into an actual `Path`.
  if isinstance(data, str):
    data = pathlib.Path(data)
  assert isinstance(data, (pd.DataFrame, pathlib.Path))

  # Check that we're dealing with a file.
  if isinstance(data, pathlib.Path):
    assert data.is_file()

  # No schema? Then generate it.
  if schema is None:
    schema = schema_inference.schema_utils.generate_schema(data, nrows)
  elif isinstance(schema, str):
    schema = utils._read_json(schema)

  # Dump `schema`.
  utils.dump_json_data(data, schema, 'schema', prefix)

  # If no functions provided, drill them.
  if functions is None:
    functions = train(data, prefix=prefix, model_types=model_types)

  # Dump `functions`.
  utils.dump_json_data(data, functions, 'driller', prefix)

  # If not custom model types specified, take'em all.
  if model_types is None:
    model_types = utils.gather_all_models(functions)
  else:
    model_types = utils.select_models(model_types, functions)
  model_types = list(map(utils.ModelType, model_types))

  # Estimate the column sizes.
  estimated_sizes = v_optimizer.compute_target_sizes(data, functions, schema, model_types)

  # Dump `estimated_sizes`.
  utils.dump_json_data(data, estimated_sizes, 'estimates', prefix)

  # Optimize the layout of the virtual table based on the previous estimates.
  functions = v_optimizer.optimize(estimated_sizes, model_types)

  # Dump `layout`.
  utils.dump_json_data(data, functions, 'layout', prefix)

  # Compress the file via the optimized layout.
  virtual_path = v_compressor.compress(data, schema, functions, optimize_layout=True, nrows=nrows)

  # Insert the functions into the metadata. They're later used when we want to decompress / query the file.
  if format_type == 'parquet':
    utils.add_metadata_pq(virtual_path, functions, 'functions')

    # Insert the schema into the metadata. TODO: Add when it's used.
    utils.add_metadata_pq(virtual_path, schema, 'schema')

    # Rename with the wished path, `format_path`.
    utils.rename(virtual_path, format_path)
  elif format_type == 'csv':
    # TODO: @Ping.
    utils.to_csv(virtual_path, format_path)
    folder_path = os.path.dirname(os.path.abspath(format_path))
    base_name = os.path.splitext(os.path.basename(format_path))[0]
    utils._write_json(os.path.join(folder_path, f'{base_name}-functions.json'), functions)
    utils._write_json(os.path.join(folder_path, f'{base_name}-schema.json'), schema)
  elif format_type == 'btrblocks':
    # TODO: @Ping: `btrblocks`.
    utils.to_btrblocks(virtual_path, format_path)
    utils._write_json(os.path.join(format_path, 'functions.json'), functions)
    utils._write_json(os.path.join(format_path, 'schema.json'), schema)  
  return

def from_format(parquet_path, functions=None, schema=None):
  # TODO: @Ping.
  pass

def query(sql_query, functions=None, schema=None, engine='duckdb', fancy=True):
  """
    Query the Parquet file via SQL.

    Args:
      sql_query (str): The SQL string. It should contain a `read_parquet(...)` call, so that we call the engine.
      functions (dict): The functions to be used during query. If `None`, then `virtual` recovers them from the Parquet's metadata.
      schema (dict): The schema of the data. If `None`, then `virtual` recovers it from Parquet's metadata.
      engine (str): The query engine to use. Options: ['duckdb', 'pandas'].
      fancy (bool): If we should also keep the original column names of the result.

    Example (1):
      virtual.query(
        'select avg(price) from read_parquet("file.parquet") where year >= 2024',
        engine = 'duckdb'
      )
  """

  # Extract the parquet path. Supports both type of quotes.
  match = re.search(r'(read_parquet|read_csv)\((["\'])(.*?)\2\)', sql_query)

  # Fallback for queries where there is no read from a file.
  if match is None:
    con = utils.get_duckdb_conn()
    ans = con.execute(sql_query).fetchdf()
    con.close()
    return ans

  # Infer the format type and the file path.
  format_type = match.group(1).replace('read_', '') if match else None
  file_path = match.group(3) if match else None

  # Check if we support this format.
  if format_type not in ['parquet']:
    assert False, f'Format {format_type} not yet supported!'

  # Read the schema.
  if schema is None:
    if format_type == 'parquet':
      schema = get_metadata_from_file(format_type, file_path, 'schema')
    else:
      assert 0, f'File format {format_type} not supported yet.'
  elif isinstance(schema, str):
    schema = utils._read_json(schema)
  assert schema is not None

  # Read the functions.
  if functions is None:
    functions = get_metadata_from_file(format_type, file_path, 'functions')
  elif isinstance(functions, str):
    functions = utils._read_json(functions)
  assert functions is not None

  # Check the query engine is valid.
  assert engine in ['duckdb', 'pandas']

  # Workaround to support `pandas`.
  if engine == 'pandas':
    if format_type == 'parquet':
      df = pd.read_parquet(file_path)
      sql_query.replace(f"read_parquet('{file_path}')", "df")
    elif format_type == 'csv':
      # TODO: @Ping.
      pass

  # Open the connection.
  con = utils.get_duckdb_conn()

  # Fetch the parquet header.
  parquet_header = utils._get_parquet_header(con, file_path)

  # TODO: Update the naming convention in the layout file to avoid this `greedy` and `chosen`.
  matches = []

  def add_match(m, raw_formula):
    # TODO: What about unicodes? The length might be different for case insensitive letters?
    matches.append((m.start(), m.start() + len(m.group()), m.group(), f'({raw_formula})'))

  for iter in functions['greedy']['chosen']:
    # Generate the formula.
    _, raw_formula = v_reconstructor.generate_formula(schema, parquet_header, iter, functions['greedy']['chosen'], enforce_replacement=True)

    pattern = re.compile(rf"\"?\b{re.escape(iter['target_name'])}\b\"?")

    # NOTE: If we have double quotes, then we *must* enforce case sensitiveness.
    for m in pattern.finditer(sql_query, re.IGNORECASE):
      # Double-quote?
      if m.group()[0].startswith('"'):
        # Should be the same, but without the double quotes (if any).
        if m.group().strip('"') == iter['target_name'].strip('"'):
          add_match(m, raw_formula)
      else:
        add_match(m, raw_formula)

  # Now comes the most clever part:
  # Since we don't want to purely replace the attributes naively (since we could replace multiple times),
  # we sort the starting positions in *decreasing* order and start replacing backwards,
  # enforcing that we search until the last token replaced. Brrr.
        
  sorted_matches = sorted(matches, key=lambda x: x[0], reverse=True)
  for start, end, before, after in sorted_matches:
    sql_query = sql_query[:start] + after + sql_query[end:]

  # Execute.
  ans = con.execute(sql_query).fetchdf()

  # Rename the columns to maintain consistency with the default output by pandas.
  if fancy:
    col_dict = dict()
    for col in ans.columns:
      new_col = col
      for _, _, before, after in sorted_matches:
        new_col = re.sub(re.escape(after), before, new_col)
      col_dict[col] = new_col
    ans = ans.rename(columns=col_dict)

  # And return.
  return ans

def functions_to_json(functions, json_path):
  utils._write_json(json_path, functions)

def get_functions_from_json(json_path):
  return utils._read_json(json_path)

def get_metadata_from_file(format_type, file_path, metadata_type):
  if format_type == 'parquet':
    return utils.get_metadata_pq(file_path, metadata_type)
  elif format_type == 'csv':
    # TODO: @Ping.
    pass
  elif format_type == 'btrblocks':
    # TODO: @Ping.
    pass
  assert 0, f'File format {format_type} not supported yet.'


import time
import duckdb
import numpy as np
import utils
from utils import ModelType
import expr_parser, expr_optimizer
from v_size import _create_regression

def collect_required_virtual_columns(target_iter, model_type, all_virtual_columns):
  cols = []
  if not model_type.is_k_regression():
    cols = utils.collect_lr_column_names(target_iter[model_type.name])
  else:
    for local_model in target_iter[model_type.name]['config']:
      cols.extend(utils.collect_lr_column_names(local_model))
  return cols

def generate_formula(schema, parquet_header, target_iter, all_virtual_columns, enforce_replacement=False):
  # Fetch the model type.
  assert 'model_type' in target_iter
  model_type = ModelType(target_iter['model_type'])

  # Determine which virtual columns we require to build extra.
  required = collect_required_virtual_columns(target_iter, model_type, all_virtual_columns)
  assert target_iter['target_name'] not in required

  # Create the mapping from the name to their expression.
  mapping = dict()
  for other_iter in all_virtual_columns:
    if other_iter['target_name'] in required:
      _, expr = generate_formula(schema, parquet_header, other_iter, all_virtual_columns, enforce_replacement=enforce_replacement)
      mapping[other_iter['target_name']] = expr

  # Declare the columns we will be using.
  # TODO: Add switch.
  is_null_col = f"{target_iter['target_name']}_null"
  outlier_col = f"{target_iter['target_name']}_outlier"
  offset_col = f"{target_iter['target_name']}_offset"
  # switch_col = f"\"{target_iter['target_name']}_switch\""

  # Simple regression.
  if not model_type.is_k_regression():
    formula = _create_regression(target_iter[model_type.name], schema, target_iter['target_name'])
  else:
    local_fs = []
    for index, local_model in enumerate(target_iter[model_type.name]['config']):
      local_fs.append(_create_regression(local_model, schema, target_iter['target_name']))

    builder = []
    for index, local_model in enumerate(target_iter[model_type.name]['config']):
      # TODO: This offset can be signed, so it's pretty bad for the type.
      if index != len(target_iter[model_type.name]['config']) - 1:
        builder.append(f"when \"{target_iter['target_name']}_switch\" = {index} then {local_fs[index]}")
      else:
        # Optimize the last `when` since we can directly put `else`.
        builder.append(f"else {local_fs[index]}")
    formula = f'case ' + '\n'.join(builder) + ' end'

  # Do we have an offset column?
  # TODO: If the offset column is a constant, we have to optimize it out.
  # TODO: Or maybe just remove it before hand within the optimizer.
  # TODO: Since we anyway do this check if we have an offset column.
  if offset_col in parquet_header:
    assert formula and formula is not None
    formula = formula + f" + \"{offset_col}\""

  # Do we have an outlier column?
  if outlier_col in parquet_header:
    formula = f"coalesce (\"{outlier_col}\", {formula})"
  
  # Do we have an `is_null` column?
  if is_null_col in parquet_header:
    formula = f"case when \"{is_null_col}\" then null else {formula} end"

  # If we really need to enforce the replacement of the required virtual columns.
  if enforce_replacement:
    for elem in mapping:
      formula = formula.replace(f"\"{elem}\"", f"({mapping[elem]})")

  # Optimize the formula.
  new_formula = None
  if not model_type.is_k_regression():
    parser = expr_parser.ExprParser(formula)
    formula_tree = parser.parse()
    optimized_tree = expr_optimizer.optimize(formula_tree)
    new_formula = expr_parser.print_expr(optimized_tree)
  else:
    # TODO: Implement the optimization for k-regression.
    new_formula = formula
  assert new_formula is not None

  # Strip out the '(' and ')'.
  if new_formula[0] == '(' and new_formula[-1] == ')':
    new_formula = new_formula[1:-1]

  # Update the formula with the optimized one.
  formula = new_formula

  # And return it, also with a suffix with the actual column name.
  return f"{formula} as \"{target_iter['target_name']}\"", formula

def _measure_impl(bench_col, parquet_path):
  # Open the connection.
  local_con = duckdb.connect(database=':memory:', read_only=False)

  # Build the sql query.
  assert bench_col is not None
  sql_query = f"select sum({bench_col}) as sum_ from read_parquet('{parquet_path}');" #  + (f" where {filter}" if filter else "") + ";"

  lats = []
  ret = None

  # NOTE: We do a warm-up query to avoid measuring the disk read.
  for _ in range(11):
    start = time.time_ns()
    try:
      # NOTE: We also fetch the result to be sure.
      ret = local_con.execute(sql_query).fetchdf()["sum_"][0]
    except Exception as e:
      print(e)
      return None

    end = time.time_ns()
    lats.append((end - start) / 1_000_000)

  # Close the connection.
  local_con.close()

  # And return.
  assert ret is not None
  return np.mean(np.asarray(lats[1:]))

def measure_reconstruction_latency_per_column(parquet_type, parquet_path, schema, all_target_columns, target_iter):
  bench_col = None
  if parquet_type == 'base':
    bench_col = f"\"{target_iter['target_name']}\""
  elif parquet_type == 'virtual':
    # Open a temporary connection.
    # This is used to read the header of the parquet file.
    local_con = duckdb.connect(database=':memory:', read_only=False)
    parquet_header = utils._get_parquet_header(local_con, parquet_path)
    local_con.close()

    # Generate the formula.
    _, raw_formula = generate_formula(schema, parquet_header, target_iter, all_target_columns, enforce_replacement=True)
    bench_col = raw_formula
  else:
    # Type `parquet_type` not supported.
    assert 0
    
  assert bench_col is not None
  return _measure_impl(bench_col, parquet_path)


def optimize(tree):
  # print(tree)

  if tree['type'] == 'cast':
    tree['expr'] = optimize(tree['expr'])
    return tree

  if tree['type'] == 'case':
    tree['when'] = list(map(optimize, tree['when']))
    tree['else'] = optimize(tree['else'])
    return tree

  if tree['type'] == 'when':
    tree['cond'] = optimize(tree['cond'])
    tree['then'] = optimize(tree['then'])
    return tree

  if tree['type'] == 'coalesce':
    tree['exprs'] = list(map(optimize, tree['exprs']))
    return tree

  if tree['type'] == 'round':
    tree['expr'] = optimize(tree['expr'])
    return tree

  if tree['type'] == 'sum':
    # First optimize the terms.
    tree['terms'] = list(map(optimize, tree['terms']))

    # Collect the local when-conditions.
    local_when_conds = []
    for term in tree['terms']:
      if term['type'] != 'term':
        continue
      if term['expr']['type'] == 'case':
        assert len(term['expr']['when']) == 1
        local_when_conds.append(term['expr']['when'][0]['cond'])

    # Nothing to optimize?
    if not local_when_conds:
      return tree

    # Merge the local when conditions.
    merged_when_conds = {
      'type' : 'or',
      'factors' : local_when_conds
    }

    # Collect the updated terms.
    updated_terms = []
    for term in tree['terms']:
      if term['type'] != 'term':
        updated_terms.append(term)
        continue
      if term['expr']['type'] != 'case':
        updated_terms.append(term)
        continue

      # Special case.
      # NOTE: We also optimize to be sure the coefficients are propagated, if any.
      # NOTE: Otherwise, we could also do this by hand, since there are at most 2 levels.
      new_stmt = optimize({
        'type' : 'term',
        'coef' : term['coef'],
        'expr' : term['expr']['else']
      })
      updated_terms.append(new_stmt)

    # TODO: Optimize the remaining terms.

    new_tree = {
      'type' : 'case',
      'when' : [{
        'type' : 'when',
        'cond' : merged_when_conds,
        'then' : {'type': 'literal', 'val': 'null'}
      }],
      'else' : {
        'type' : 'sum',
        'terms' : updated_terms
      }
    }

    return new_tree

  if tree['type'] == 'term':
    tree['expr'] = optimize(tree['expr'])
    if tree['expr']['type'] == 'term':
      assert 'coef' in tree['expr']
      merged_coef = tree['coef'] * tree['expr']['coef']
      return {
        'type' : 'term',
        'coef' : merged_coef,
        'expr' : tree['expr']['expr']
      }
    return tree

  if tree['type'] == 'literal':
    return tree

  assert 0


import sys
import os

DEBUG_MODE = False

def _block_print():
  sys.stdout = open(os.devnull, 'w')

def _enable_print():
  sys.stdout = sys.__stdout__

def _check_debug_mode():
  if DEBUG_MODE:
    _enable_print()
  else:
    _block_print()




import duckdb
import os
import pandas as pd
import pyarrow.parquet as pq
from typing import List
import utils
from utils import ModelType, _get_info, _get_nullable, _get_column, get_data_hash
import math
import pathlib
from typing import Optional

PARQUET_COMPRESSION_TYPE = 'snappy'

def arrow_compress_parquet(in_file, out_file):
  table = pq.read_table(in_file)
  pq.write_table(table, out_file, compression=PARQUET_COMPRESSION_TYPE)

def _model_needs_offset(iter, model_type):
  return True
  # return iter["metrics"][model_type]['equal'] != 1.0

# TODO: Do we really need this?
def _model_has_null_reference_columns(schema, model, model_type: ModelType):
  # Is the model a k-regression?
  if model_type.is_k_regression():
    # TODO: What if that model is never selected.
    # TODO: I mean, if that's the case, we should directly drop it in the explorer, right?
    for local_model in model['config']:
      for iter in local_model['coeffs']:
        if _get_column(schema, iter['col_name'])['null']['any']:
          return True
  else:
    for iter in model['coeffs']:
      if _get_column(schema, iter['col_name'])['null']['any']:
        return True
  return False

def _is_target_column_null(schema, target_column):
  return _get_column(schema, target_column['target_name'])['null']['any']

def get_model(iter, model_type: ModelType):
  # Have we already assigned a model type to this column?
  if 'model_type' in iter:
    # If so, just return the actual model.
    return iter[iter['model_type']], ModelType(iter['model_type'])
  else:
    # Otherwise, we must have a custom `model_type`.
    # In that case, check if it is present as a model.
    if 'models' in iter:
        # Nothing found?
        if model_type.name not in iter['models']:
          return None, model_type
        
        # Otherwise, return the model.
        return iter['models'][model_type.name], model_type
    
  # This should never happen.
  assert 0
  return None, None

def _create_base_table(target_columns, schema, model_type=None):
  columns = []
  for iter in schema:
    type = _get_info(schema, iter['name'])['type']
    nullable = _get_nullable(schema, iter['name'])
    columns.append(f'"{iter["name"]}" {type} {nullable}')

  def create_offset_column(target_name):
    name = f"{target_name}_offset"
    type = _get_info(schema, target_name)['type']
    return f"\"{name}\" {type} not null default 0"

  def create_outlier_column(target_name):
    name = f"{target_name}_outlier"
    type = _get_info(schema, target_name)['type']
    return f"\"{name}\" {type} null default null"

  def create_is_null_column(target_name):
    # NOTE: Only change this if you know what to do.
    # We use split('_') later and this could destroy the functionality of the script.
    name = f"{target_name}_null"
    return f"\"{name}\" bool not null default false"

  def create_switch_column(target_name):
    name = f"{target_name}_switch"
    # TODO: You need to restore to `not null`.
    # TODO: There was a bug in taxi, that's why we disabled it.
    return f"\"{name}\" smallint not null default 0"

  def create_formula_column(target_name, index):
    name = f"{target_name}_f{index}"

    # TODO: Actually, to avoid any problems, we should make this as large as it gets.
    type = _get_info(schema, target_name)['type']
    return f"\"{name}\" {type}"

  def create_least_column(target_name):
    name = f"{target_name}_least"
    type = _get_info(schema, target_name)['type']

    # TODO: This part should be signed, I think.
    # Or not, since we do abs.
    # But the operation inside that should be casted as double, right???
    return f"\"{name}\" {type}"

  # def create_temp_offset_column(target_name):
  #    name = f"{target_name}_temp_offset"
  #    type = _get_info(schema, target_name)['type']
  #    return f"\"{name}\" {type} default 0"

  # TODO: Maybe we can later remove these columns if they're not used.
  for iter in target_columns:
    # Fetch the model.
    model, model_type = get_model(iter, model_type)

    # No model? This can only happen if the `model_type` was specified before hand. 
    if model is None:
      assert model_type is not None
      continue

    # Do we need an `offset` column?
    # NOTE: Currently, we always create one. We remove it afterwards in an optimization step.
    if _model_needs_offset(iter, model_type):
      columns.append(create_offset_column(iter['target_name']))

    # Do we need an `outlier` column? This happens if one of the reference columns has NULLs.
    if _model_has_null_reference_columns(schema, model, model_type):
      columns.append(create_outlier_column(iter['target_name']))

    # Do we need an `null` column? This happens if the column has NULLs.
    if _is_target_column_null(schema, iter):
      columns.append(create_is_null_column(iter['target_name']))

    # Add the switch column.
    if model_type.is_k_regression():
        # TODO: Maybe don't do this if we actually don't find anything.
        # TODO: But this can already be done in the explorer.
        columns.append(create_switch_column(iter['target_name']))

    # This part only concerns temporary columns.
    if model_type.is_k_regression():
        # TODO: The `k` should be optimized in the explorer.
        k = model_type.extract_k()

        for index in range(k):
          columns.append(create_formula_column(iter['target_name'], index))
        
        # Add the `least` column.
        columns.append(create_least_column(iter['target_name']))

    # Add a temporary offset column which can be null.
    # if utils.is_k_regression(model_type):
    #    columns.append(create_temp_offset_column(iter['target_name']))

  columns = " , ".join(columns)

  return f"CREATE TABLE base_table ({columns});"

def _create_sample(data: pd.DataFrame | pathlib.Path, sample_path, sample_size):
  # If there is only a path, first read it.
  if isinstance(data, pathlib.Path):
    # TODO: We should ensure there is no header!
    df = pd.read_csv(data, dtype=str, nrows=sample_size, header=None)
  elif isinstance(data, pd.DataFrame):
    # Otherwise, use the dataframe we already have.
    if sample_size < len(data):
      df = data.iloc[:sample_size]
    else:
      df = data
  
  # Dump to `sample_path`.
  df.to_csv(sample_path, index=False, header=False)

def create_copy_csv(data: pd.DataFrame | pathlib.Path, schema, nrows=None):
  assert isinstance(data, (pd.DataFrame, pathlib.Path))

  # Take all columns we have to read.
  columns = ','.join([f'"{iter["name"]}"' for iter in schema])

  # Path?
  if isinstance(data, pathlib.Path):
    # Infer the delimiter.
    dialect = utils.infer_dialect(data)

    # Create the load SQL.
    # TODO: I now know why it works with the estimator and not with this one.
    # TODO: We always created the sample that uses `_create_sample` (pandas.df), which probably removed the NULLs.
    # TODO: We can do this better, really.
    load_sql = f"insert into base_table({columns}) select * from read_csv('{data}', header=true, delim='{dialect['delimiter']}', quote='{dialect['quotechar']}'" \
            + (f", sample_size={nrows}" if nrows is not None else "") \
            + ", nullstr=['NULL', 'null', '', 'None'])"
  elif isinstance(data, pd.DataFrame):
    # Then we have a dataframe.
    # TODO: Do we still need this one?
    # TODO: Should we actually say `TIMESTAMP` instead of `timestamp`?
    # TODO: Wait -- we might actually modify user's data?
    for i in schema:
      if i['type'] == 'timestamp':
        data[i['name']] = pd.to_datetime(data[i['name']])
    load_sql = f'insert into base_table({columns}) select * from data'
  else:
    assert 0

  # Add the LIMIT clause to bound until `nrows`.
  if nrows is not None:
    load_sql = f'{load_sql} limit {nrows}'

  # And return the SQL query.
  return load_sql

# Create a regression model.
def _create_regression(model, schema, target_name):
  # This will contain the regression components.
  regression = []

  # Get the information about the column.
  info = _get_info(schema, target_name)

  formula_has_only_integers = True
  all_coeffs_are_integers = True

  scales = []
  for iter in model['coeffs']:
    # Is the coefficient 0?
    # Then skip this part.
    if math.isclose(iter['coeff'], 0.0):
      continue

    # TODO: What if the coefficient is 0? We could skip this.
    # TODO: But then we also need to adapt this for the `is null` part.
    # TODO: This is wrong.
    if not utils.is_integer(_get_info(schema, iter['col_name'])['type']):
      formula_has_only_integers = False

    # Take the current scale.
    scales.append(_get_info(schema, iter['col_name'])['scale'])

    # Optimize out a 1.0 coefficient.
    if math.isclose(iter['coeff'], 1.0):
      regression.append(f"\"{iter['col_name']}\"")
    else:
      rounded_coeff = utils.custom_round(iter['coeff'])

      # Check if it is an integer.
      if not rounded_coeff.is_integer():
        formula_has_only_integers = False
        all_coeffs_are_integers = False
      else:
        rounded_coeff = int(rounded_coeff)
      regression.append(f"{rounded_coeff} * \"{iter['col_name']}\"")

  # Analyze the intercept.
  intercept = utils.custom_round(model['intercept'])

  # Not 0?
  if not math.isclose(intercept, 0.0):
    # TODO: Should we also round this?
    if not intercept.is_integer():
      all_coeffs_are_integers = False
      formula_has_only_integers = False
    else:
      # Otherwise make it a true integer.
      intercept = int(intercept)
    regression.append(str(intercept))
  elif len(regression) == 0:
    # NOTE: This is really a corner case. I am really glad we caught it.
    # This happens if everything is 0.
    regression.append(str(0))

  # And build the formula.
  formula = ' + '.join(regression)

  # Is the target column an integer?
  # Then force the offset to also be one.
  # TODO: Note, this can also overflow.
  if utils.is_integer(info['type']):
    actual_type = info['type']

    # Only round if there was a different type than integer in the formula.
    # Otherwise, let is as is.
    # TODO: Probably problems with overflow?
    if not formula_has_only_integers:
      formula = f'round({formula}, 0)::{actual_type}'
  elif utils.is_double(info['type']):
    scale = info['scale']

    # If we only have this scale.
    # TODO: Maybe even with lower scales would be better?
    # TODO: For instance, if we have integers.
    # EXAMPLE: [2, 2, 0, 1], and we have the scale = 2, than this would work.
    # So if scale >= max(scale).

    # Corner case: There is only the intercept in the regression.
    if not len(model['coeffs']):
      assert not len(scales)
      assert len(regression) == 1
      pass
    elif scale >= max(scales) and all_coeffs_are_integers:
      # NOTE: This only if all coefficients are integers.
      # Otherwise, we need to cast.
      pass
    else:
      # TODO: Maybe this is the right place to cast to the actual type.
      # TODO: But we keep DOUBLE for now.
      formula = f'round({formula}, {scale})'

  # And return.
  return formula

def _create_abs_offset(formula, col):
  return f"abs(\"{col}\" - {formula})"

def create_dump_base(con, schema, target_columns, out_file):
  columns = ",".join([f'"{iter["name"]}"' for iter in schema])
  return f"copy (select {columns} from base_table) to '{out_file}' (FORMAT PARQUET, COMPRESSION '{PARQUET_COMPRESSION_TYPE}');"

def create_dump_base_target_column(con, schema, target_column, out_file):
  return f"copy (select \"{target_column['target_name']}\" from base_table) to '{out_file}' (FORMAT PARQUET, COMPRESSION '{PARQUET_COMPRESSION_TYPE}');"

# NOTE: You need to adapt this part if you want change the layout.
def is_auxiliary_column(cn):
  token = cn.split('_')[-1].strip('"')
  return utils.is_formula(token) or token == 'least'

def is_special_column(cn):
  token = cn.split('_')[-1].strip('"')
  return token in ['switch', 'offset', 'outlier', 'null']

def clean_base_table(all_columns, target_columns):
  rest = all_columns.copy()
  for target_column in target_columns:
    for cn in all_columns:
      # TODO: We also need to take into account "...", right?
      # TODO: This is really a hack.
      # NOTE: Pay attention at this case: average_admissions_30_49_covid_confirmed_per_100k vs average_admissions_30_49_covid_confirmed.
      if utils.is_attached_column_of(target_column['target_name'], cn) and is_auxiliary_column(cn):
        rest.remove(cn)
  return rest

def create_dump_virtual(con, schema, target_columns, out_file):
  # Get the current header.
  all_columns = utils._get_curr_columns(con)

  # Clean the table.
  all_columns = clean_base_table(all_columns, target_columns) 

  # Remove the original target column names.
  # TODO: We should take care of spaces in the column names, right?
  for iter in target_columns:
    all_columns.remove(iter["target_name"])

  # And flush to parquet.
  all_columns = [f"\"{iter}\"" for iter in all_columns]
  all_columns = ",".join(all_columns)
  return f"copy (select {all_columns} from base_table) to '{out_file}' (FORMAT PARQUET, COMPRESSION '{PARQUET_COMPRESSION_TYPE}');"

def create_dump_virtual_target_column(con, schema, target_column, out_file):
  # Get the current header.
  all_columns = utils._get_curr_columns(con)

  # Clean the table.
  all_columns = clean_base_table(all_columns, [target_column]) 

  # Remove the target column.
  all_columns.remove(target_column['target_name'])

  # Take the attached column.
  keep = []
  for cn in all_columns:
    # TODO: Do we need to use quotes?
    if utils.is_attached_column_of(target_column['target_name'], cn):
        keep.append(cn)
  
  # If there is nothing to keep, just store a dummy parquet file.
  # TODO: Note - if you are to read this, it becomes complicated.
  select_clause = '1 as dummy_column'
  if keep:
    keep = [f"\"{iter}\"" for iter in keep]
    select_clause = ",".join(keep)

  # And flush to parquet.
  return f"copy (select {select_clause} from base_table) to '{out_file}' (FORMAT PARQUET, COMPRESSION '{PARQUET_COMPRESSION_TYPE}');"

def build_least_column(model, schema, target_iter):
  least_builder = []
  updates = []
  for index, local_model in enumerate(model['config']):
    # NOTE: This formula has already taken into consideration the type of the target column.
    local_formula = _create_regression(local_model, schema, target_iter['target_name'])

    # Build the `least` column.
    # TODO: The problem is that the difference could be negative.
    # TODO: Or maybe we do an if.
    # TODO: Done, but this could still lead to an overflow.
    local_offset = _create_abs_offset(local_formula, target_iter['target_name'])
    least_builder.append(local_offset)

    updates.append(f"\"{target_iter['target_name']}_f{index}\" = {local_formula}")

  # Build.
  least_builder_str = ', '.join(least_builder)
  least_builder_str = f"\"{target_iter['target_name']}_least\" = least({least_builder_str})"

  updates.append(least_builder_str)

  # And return.
  return f"update base_table set {', '.join(updates)};"

def build_switch_column(model, schema, target_iter):
  switch_builder = []
  for index, local_model in enumerate(model['config']):
    local_formula = _create_regression(local_model, schema, target_iter['target_name'])
    local_offset = _create_abs_offset(local_formula, target_iter['target_name'])
    switch_builder.append(f"when {local_offset} - \"{target_iter['target_name']}_least\" < 1e-6 then {index}")
  
  # Build. Please keep the space before the first ` end`, otherwise we have problems.
  switch_builder_str = f"case when \"{target_iter['target_name']}_least\" is null then 0 else (case " + '\n'.join(switch_builder) + " end) end"
  
  # And return.
  return f"update base_table set \"{target_iter['target_name']}_switch\" = {switch_builder_str};"

def build_offset_column(model, schema, target_iter):
  # TODO: Maybe we can put the switch to -1, and then update it to 0.
  offset_builder = []
  for index, local_model in enumerate(model['config']):
    local_formula = _create_regression(local_model, schema, target_iter['target_name'])
    
    # TODO: This offset can be signed, so it's pretty bad for the type.
    offset_builder.append(f"when \"{target_iter['target_name']}_switch\" = {index} then \"{target_iter['target_name']}\" - {local_formula}")
  
  # Build.
  offset_builder_str = f"case when \"{target_iter['target_name']}_least\" is null then 0 else (case " + '\n'.join(offset_builder) + " end) end"
  
  # And return.
  return f"update base_table set \"{target_iter['target_name']}_offset\" = {offset_builder_str};"

def debug_base_table(con, msg, verbose=False):
  if verbose:
    print(f'\n ---- @@@ {msg} @@@ ---- \n')
    tmp = con.execute(f'select * from base_table;').fetchdf()
    print(tmp[tmp.columns.sort_values()].head(100).to_markdown())
  pass

def create_virtual_column_layout(con, target_iter, schema, model_type: Optional[ModelType]=None):
  def _create_sql_any_reference_column_is_null(model):
    # TODO: But this is _not_ the point, since we just want to check if it is null.
    # TODO: Not that this is nullable.
    # TODO: Even in that case - why should be the switch column NULL when all are nulls?
    reference_columns = [it for it in model["coeffs"] if _get_column(schema, it["col_name"])["null"]]
    reference_columns = [f"\"{it['col_name']}\" is null" for it in reference_columns]
    reference_columns = ' or '.join(reference_columns)

    # If we really do not have any nullable columns, just return `false` (the boolean SQL type).
    if not reference_columns:
      return 'false'
    
    # Otherwise, return the default.
    return reference_columns

  # In case the column has _only_ NULLs, skip.
  if _get_column(schema, target_iter['target_name'])['null']['all']:
    return []

  # Fetch the model.
  model, model_type = get_model(target_iter, model_type)
  if model is None:
    return []

  # The result.
  local_updates = []

  # Add temporary columns in case of a k-regression.
  if model_type.is_k_regression():      
    try:
      # Build the least column.
      con.execute(build_least_column(model, schema, target_iter))

      debug_base_table(con, 'after least')
      
      # Build the switch column.
      con.execute(build_switch_column(model, schema, target_iter))

      debug_base_table(con, 'after switch')
      
      # Build the offset column.
      con.execute(build_offset_column(model, schema, target_iter))

      debug_base_table(con, 'after offset')
    except Exception as e:
      # TODO: Just a hack to avoid other issues.
      # TODO: This fails for the divvy dataset, where the intercept is too large.
      # TODO: Yet, sparse solves this much better.
      return []

  # Add offset column.
  if not model_type.is_k_regression():
    # TODO: There is a problem here, since the offset doesn't take into account the outlier.
    if _model_needs_offset(target_iter, model_type):
      # Create the regression.
      formula = _create_regression(model, schema, target_iter['target_name'])

      # NOTE | TODO: The offset should also take the scale of the original column?
      update = f"round(\"{target_iter['target_name']}\" - ({formula}), {utils._get_info(schema, target_iter['target_name'])['scale']})"
      update = f"coalesce({update}, 0)"
      update = f"\"{target_iter['target_name']}_offset\" = {update}"
      local_updates.append(update)
  
  # Check for null reference columns.
  if _model_has_null_reference_columns(schema, model, model_type):
    # Simple regression?
    expr = None
    if not model_type.is_k_regression():
      ref_is_null = _create_sql_any_reference_column_is_null(model)
      expr = f"case when {ref_is_null} then \"{target_iter['target_name']}\" else null end"
    else:
      # At this point, we need to only take _our_ reference columns, i.e., those corresponding to the switch.         
      # Build the expressions for the `is null` expressions.
      ref_is_null = []
      for it in model['config']:
        ref_is_null.append(_create_sql_any_reference_column_is_null(it))

      # Build the outlier column.
      outlier_builder = []
      for index in range(len(model['config'])):
        col = target_iter['target_name']

        # Inspect the switch.            
        outlier_builder.append(f"when \"{col}_switch\" = {index} then (case when {ref_is_null[index]} then null else \"{col}\" end)")
      expr = f'case ' + '\n'.join(outlier_builder) + ' end'

    # Set the outlier column.
    assert expr is not None
    local_updates.append(f"\"{target_iter['target_name']}_outlier\" = {expr}")

  # Check if the column can be null.
  if _is_target_column_null(schema, target_iter):
    update = f"\"{target_iter['target_name']}_null\" = \"{target_iter['target_name']}\" is null"
    local_updates.append(update)

  # And return.
  return local_updates

# TODO: This is the one with the issue.
def create_virtual_table_layout(con, target_columns, schema, model_type=None):
  # Virtualize each possible column.
  for iter in target_columns:
    curr_updates = create_virtual_column_layout(con, iter, schema, model_type)

    # No updates? Then skip.
    if not curr_updates:
      continue

    # Try to update it.
    try:
      update_sql = f"update base_table set {','.join(curr_updates)};"
      con.execute(update_sql)

      debug_base_table(con, f"{iter['target_name']} @@@@@ after the updates!!!!!!!!!!")
    except Exception as e:
      # Otherwise, skip it.
      continue
  return

def reoptimize_virtual_table_layout(con, target_columns):
  freqs = utils._get_distinct_counts(con)

  to_remove = []
  for target_column in target_columns:
    for cn in freqs:
      # Is it actually distinct?
      if freqs[cn][0] != 1:
        continue

      assert cn.endswith('_count')
      cn = utils.rreplace(cn, '_count', '')

      # NOTE: This is really important!
      if cn == target_column['target_name']:
        continue

      # Then remove it if it is an auxiliary column.
      if utils.is_attached_column_of(target_column['target_name'], cn) and is_special_column(cn):
        to_remove.append(cn)

  if len(to_remove):
    for cn in to_remove:
      con.execute(f"alter table base_table drop column \"{cn}\";")
  pass

def _size_impl(con, schema, target_column, model_type=None, data_hash=None):
  assert data_hash is not None

  def get_size(dump_callback, target_column):
    out_file = "size"

    sql = dump_callback(con, schema, target_column, f'.{out_file}_duckdb-{data_hash}.parquet')
    if len(sql) == 0:
      return 0

    # Execute.
    con.execute(sql)

    # Compress via arrow.
    arrow_compress_parquet(f'.{out_file}_duckdb-{data_hash}.parquet', f'.{out_file}_arrow-{data_hash}.parquet')

    # Calculate the final size.
    size_to_return = os.path.getsize(f'.{out_file}_arrow-{data_hash}.parquet')

    # Remove both parquet files, since they are just temporary.
    os.remove(f'.{out_file}_duckdb-{data_hash}.parquet')
    os.remove(f'.{out_file}_arrow-{data_hash}.parquet')

    # And return the size.
    return size_to_return

  # Fetch the model.
  model, model_type = get_model(target_column, model_type)
  if model is None:
    return

  # The raw size of the base column.
  base_target_column_size = get_size(create_dump_base_target_column, target_column)

  # Its virtual size.
  virtual_target_column_size = get_size(create_dump_virtual_target_column, target_column)
  
  # Add the sizes into the dictionary.
  if 'sizes' not in target_column:
    target_column['sizes'] = dict()
  
  # NOTE: This is _not_ an `elif`!
  if model_type.name not in target_column['sizes']:
    target_column['sizes'][model_type.name] = dict()

  # And add them.
  target_column['sizes'][model_type.name]['base'] = base_target_column_size
  target_column['sizes'][model_type.name]['virtual'] = virtual_target_column_size
  return

# We use this function to estimate the {arquet sizes for the columns.
# We say "estimate", since we only analyze a sample size of 10K tuples.
def compute_sizes_of_target_columns(functions, data: pd.DataFrame | pathlib.Path, schema, model_types: List[ModelType], sample_size=10_000):
  assert isinstance(data, (pd.DataFrame, pathlib.Path))
  # TODO: Maybe we can sample only once?

  # Construct an unique hash.
  data_hash = get_data_hash(data)

  for model_type in model_types:
    # Make a new connection to be sure that we do not use the same tables as the previous model.
    con = duckdb.connect(database=':memory:', read_only=False)

    # Create the base table.
    # NOTE: This also creates the auxiliary columns (only the schema thereof).
    con.execute(_create_base_table(functions, schema, model_type))
    sample_path = f'.sample-{data_hash}.csv'
    
    debug_base_table(con, 'after create basetable')

    # Load the sample.
    _create_sample(data, sample_path, sample_size)

    # Copy the CSV into the SQL table.
    con.execute(create_copy_csv(pathlib.Path(sample_path), schema, nrows=sample_size))

    # Remove the sample.
    os.remove(sample_path)

    # Finalize the layout. This works for both simple and k-regression.
    create_virtual_table_layout(con, functions, schema, model_type)

    debug_base_table(con, 'After `create_virtual_table_layout`.')

    # Compute the base and virtual sizes.
    for target_column in functions:
      _size_impl(con, schema, target_column, model_type=model_type, data_hash=data_hash)

    # And close the connection.
    con.close()

  return functions


import os
import json
import yaml
import utils

def schema_to_yaml(schema):
  dmap = {
    "VARCHAR": "string",
    "BIGINT": "integer",
    "SMALLINT": "integer",
    "DOUBLE": "double",
    "TIMESTAMP": "string",
    "BOOLEAN": "integer",
    "DATE": "string",
    "TIME": "string"
  }
  # schema = utils._read_json(schema)
  yaml_data = {"columns": None}
  columns_tmp = []
  for col in schema:
    columns_tmp.append({"name": col["name"], "type": dmap[col["type"]]})

  yaml_data["columns"] = columns_tmp
  return yaml_data

def generate_yaml_doc(yaml_file, schema):
  yaml_data = schema_to_yaml(schema)
  file = open(yaml_file, 'w', encoding='utf-8')
  yaml.dump(yaml_data, file)
  file.close()

def convert_csv_to_btrblocks(csv_path: str, format_path: str, schema: str | list, convertor_path = '../../btrblocks/build/'):
  if isinstance(schema, str):
    schema = utils._read_json(schema)
  yaml_name = "tmp.yaml"
  extra_path = os.path.join(format_path, 'stats')
  btr_path = os.path.join(format_path, 'btr/')
  binary_path = os.path.join(format_path, 'binary/')
  yaml_path = os.path.join(extra_path, yaml_name)
  os.makedirs(format_path, exist_ok=True)
  os.makedirs(extra_path, exist_ok=True)
  # print(f"format_path {format_path}")
  # print(f"yaml_path {yaml_path}")

  generate_yaml_doc(yaml_path, schema)
  os.system(f"rm -rf {btr_path}")
  os.system(f"rm -rf {binary_path}")
  try:
    os.system(f'{os.path.join(convertor_path, 'csvtobtr')} -create_btr -csv {csv_path} -yaml {yaml_path} -binary {binary_path} -create_binary -btr {btr_path} -stats {os.path.join(extra_path, 'stats.txt')} -compressionout {os.path.join(extra_path, 'compression.txt')}')
  except:
    print(f"Error processing {csv_path}.")

  os.system(f"rm -rf {yaml_path}")
  stas = get_status(extra_path)
  os.system(f"rm -rf {binary_path}")
  return stas

def get_status(extra_path, csv_name = None):
  if csv_name is None:
    stas = {
      'total': {},
      'columns': {}
    }
  else:
    stas = {
      'filename': csv_name+'.csv',
      'total': {},
      'columns': {}
    }
  with open(f"{os.path.join(extra_path, 'stats.txt')}", 'r') as f:
    rows = f.read()
  elements = rows.split('\n')[1:-1]
  for i, row in enumerate(elements):
    if i < len(elements) - 1:
      row_e = row.rsplit(',', 2)
      stas['columns'][row_e[0].split('_', 1)[1]] = {'btrblock-uncompressed': int(row_e[1]), 'btrblock-compressed': int(row_e[2])}
    else:
      row_e = row.split(',')
      stas['total'] = {'btrblock-uncompressed': int(row_e[1]), 'btrblock-compressed': int(row_e[2])}
          
  return stas


from virtual.v_optimizer import optimize
import pytest
from virtual.utils import ModelType, _read_json

@pytest.fixture
def model_types():
  return list(map(ModelType, ['sparse-lr']))

@pytest.fixture
def sample_data1():
  # Generic example.
  return _read_json('tests/sample_data1.json')

@pytest.fixture
def sample_data2():
  # Counter-example for the greedy algorithm.
  return _read_json('tests/sample_data2.json')

def test_optimize1(sample_data1, model_types):
  ret = optimize(sample_data1, model_types)
 
  # We only take 2 functions.
  assert ret['greedy']['obj-value'] == 2

def test_optimize2(sample_data2, model_types):
  ret = optimize(sample_data2, model_types)
 
  # TODO: This could be better by using the optimal algorithm.
  # We only take 2 functions.
  assert ret['greedy']['obj-value'] == 1.5



from schema_inference.lightweight_schema_inference import LWSchemaInferer
from schema_inference.heavyweight_schema_inference import HWSchemaInferer
from virtual.utils import infer_dialect, get_default_csv_dialect
import pandas as pd
import pathlib
import re

# Generate the schema by the heavyweight variant (it takes a while if there are many columns).
def generate_schema(data: pd.DataFrame | pathlib.Path, nrows=None):
  # Convert to an actual path.
  if isinstance(data, str):
    data = pathlib.Path(data)
  
  # Build the inferer.
  inferer = HWSchemaInferer(data)

  # And return.
  return inferer.infer(nrows)

def map_sql_to_pandas(sql_type):
  type_mapping = {
    'timestamp': 'datetime64[ns]',
    'datetime' : 'datetime64[ns]',
    'date' : 'datetime64[ns]',
    'time' : 'datetime64[ns]',
    'decimal': 'float64',
    'double': 'float64',
    # TODO: Fix this later. The issue is that DuckDB interprets as `BOOLEAN` any binary column.
    'boolean': 'str',
    'varchar': 'str',
    'char' : 'str',
    # TODO: is this a problem?
    'smallint': pd.Int16Dtype(),
    'int': pd.Int32Dtype(),
    'bigint': pd.Int64Dtype()
    # Add other types as needed
  }
    
  # Extract the base SQL type (ignore size specifiers in parentheses)
  base_type = re.match(r'^\w+', sql_type.lower()).group(0)

  # If the base type is decimal, handle precision and scale
  if base_type == 'decimal':
    # Note: `pandas`` will use float64 for decimals
    return 'float64'

  return type_mapping.get(base_type, 'object')

def handle_schema(data: pd.DataFrame | pathlib.Path, nrows=None):
  dtype_mapping = {}
  column_names, date_cns, string_cns, boolean_cns = [], [], [], []

  # Read the colum types.
  inferer = LWSchemaInferer(data)

  # And infer
  schema, has_header = inferer.infer(nrows=nrows)

  # Build the type mapping.
  # TODO: I don't get why we still need this one.
  # TODO: I mean, if we'll switch to DuckDB, then there is no reason, why we should still do this.
  for col in schema:
    column_names.append(col)
    typ = schema[col]

    # TODO: There are still some hacks we did. 
    pandas_type = map_sql_to_pandas(typ)

    if pandas_type == 'datetime64[ns]':
      date_cns.append(col)
    else:
      dtype_mapping[col] = pandas_type
      if pandas_type == 'str':
        string_cns.append(col)
      elif pandas_type == 'boolean':
        boolean_cns.append(col)

  # Infer the delimiter.
  if isinstance(data, pathlib.Path):
    csv_dialect = infer_dialect(data)
  elif isinstance(data, pd.DataFrame):
    # Default.
    csv_dialect = get_default_csv_dialect()
  else:
    assert 0

  # And return.
  return column_names, dtype_mapping, csv_dialect, has_header, {
    'date' : date_cns,
    'string' : string_cns,
    'boolean' : boolean_cns
  }


import pandas as pd
import pathlib
from virtual.utils import infer_dialect
from .lightweight_schema_inference import LWSchemaInferer

class TypeChecker:
  def __init__(self):
    self.reset()
    pass

  def reset(self):
    self.left_precision, self.right_precision = 0, 0
    pass

  def _is_bool(self, value):
    try:
      return (len(value) == 4 or len(value) == 5) and value.lower() in ['true', 'false', 't', 'f']
    except:
      return False

  def _is_integer(self, value):
    # Check if any `.` in the string representation.
    if '.' in str(value):
      return False
    
    # Otherwise, try a cast.
    try:
      result = int(value)
      return result < 2**31 - 1 and result > -2**31
    except:
      return False
      
  def _is_bigint(self, value):
    # Check if any `.` in the string representation.
    if '.' in str(value):
      return False
    
    # Otherwise, try a cast.
    try:
      int(value)
      return True
    except:
      return False
      
  def _is_float(self, value):
    # NOTE: We no longer use `DecimalTuple`.
    # This is because for some numbers it even increased the number of decimals.

    # Calculate the precision.
    # NOTE: We also need to account for the possible minus sign in front of `x`.
    def calculate_precision(x):
      if x.startswith('-'):
        return len(x) - 1
      return len(x)

    try:
      # Convert to float.
      float(value)

      # In case we have an integer.
      # NOTE: This also includes the case in which we have an integer.
      if len(value.split('.')) == 1:
        # Update only the left precision.
        self.left_precision = max(self.left_precision, calculate_precision(value))
        return True

      # Assert that we always have two parts. 
      if len(value.split('.')) > 2:
        return False
      
      # Count the left and right precision.
      # NOTE: The left part could start with a negative sign, so we have to take that into account.
      self.left_precision = max(self.left_precision, calculate_precision(value.split('.')[0]))
      self.right_precision = max(self.right_precision, calculate_precision(value.split('.')[1]))
      return True
    except:
      return False

  def _is_null(self, value, verbose=False):
    if verbose:
      print(value)
    return value is None or pd.isna(value)

class HWSchemaInferer:
  def __init__(self, data: pd.DataFrame | pathlib.Path):
    assert isinstance(data, (pd.DataFrame | pathlib.Path))
    self.data = data

  def infer(self, nrows=None):
    # Infer the schema via DuckDB.
    inferer = LWSchemaInferer(self.data)
    type_mapping, _ = inferer.infer(nrows=nrows)

    # Is `data` a path?
    if isinstance(self.data, pathlib.Path):
      # TODO: Do this better.
      # Infer the csv dialect.
      csv_dialect = infer_dialect(self.data)

      # Read dataframe. NOTE: If `nrows` is `None`, then the entire file is read.
      # TODO: Replace by DuckDB?
      df = pd.read_csv(
        self.data,
        dtype=str,
        nrows=nrows,
        # If the file contains a header row, then you should explicitly pass header=0 to override the column names.
        header=0,
        names=type_mapping.keys(),
        delimiter=csv_dialect['delimiter'],
        quotechar=csv_dialect['quotechar']
      )
    elif isinstance(self.data, pd.DataFrame):
      # Do we have a number of rows specified.
      if nrows is not None:
        df = self.data.iloc[:nrows]
      else:
        # If not, then take the entire data. That's also the default in `pd.read_csv`.
        df = self.data
    else:
      # Unreachable.
      assert 0

    # A type checker.
    type_checker = TypeChecker()

    # Check if `col_name` has NULLs.
    def check_for_null(col_name):
      tmp = df[col_name].isnull()
      return {
        'any' : bool(tmp.any()),
        'all' : bool(tmp.all())
      }

    # Calculate the precision and the scale of `col_name`.
    def calculate_precision_and_scale(col_name):
      # Reset.
      type_checker.reset()

      # Iterate the column.
      for _, value in enumerate(df[col_name]):
        # Skip NULL values.
        if type_checker._is_null(value):
          continue
        assert type_checker._is_float(str(value))

      return type_checker.left_precision + type_checker.right_precision, type_checker.right_precision

    schema = []
    for col_name in type_mapping:
      col_dict = {
        # NOTE: We remove the trailing spaces so that the column name matching is more efficient in the actual code.
        'name': col_name.strip(),
        'type' : type_mapping[col_name],
        'null': check_for_null(col_name),
        'scale' : 0,
        'precision' : 0
      }
      if type_mapping[col_name] == 'DOUBLE':
        col_dict['precision'], col_dict['scale'] = calculate_precision_and_scale(col_name)

      # Add to the schema.
      schema.append(col_dict)

    # And return it.
    return schema